import jax
import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: needs human review
class EXTROSNB(AbstractUnconstrainedMinimisation):
    """Extended Rosenbrock function (nonseparable version).

    This is a scaled variant of the Rosenbrock function.
    The function is characterized by a curved narrow valley.

    The objective function is:
    f(x) = (x_1 + 1)^2 + 100 * sum_{i=2}^{n} (x_i - x_{i-1}^2)^2

    Source: problem 21 in
    J.J. Moré, B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    SIF input: Ph. Toint, Dec 1989.

    Classification: SUR2-AN-V-0
    """

    n: int = 1000  # Default dimension (other suggested dimensions: 5, 10, 100)

    def objective(self, y, args):
        del args

        # From SIF file:
        # Group SQ1 = (X1 + 1.0) with L2 group type → (X1 + 1.0)²
        # Group SQ(I) = X(I) + ELA(I) with scale 0.01 and L2 group type
        # where ELA(I) = -X(I-1)² from ETYPE element
        # So SQ(I) = 0.01 × (X(I) - X(I-1)²)² for I=2..N

        # First term: (X1 + 1.0)²
        term1 = (y[0] + 1.0) ** 2

        # Remaining terms: 0.01 × (X(I) - X(I-1)²)² for I=2..N
        def scaled_term(i):
            # i ranges from 1 to n-1 (0-based), corresponding to SIF I=2..N
            return 0.01 * (y[i] - y[i - 1] ** 2) ** 2

        indices = jnp.arange(1, self.n)
        term2 = jnp.sum(jax.vmap(scaled_term)(indices))

        return term1 + term2

    def y0(self):
        # Starting point from the SIF file: all variables = -1.0
        return jnp.full(self.n, -1.0)

    def args(self):
        return None

    def expected_result(self):
        # The optimal solution has all components equal to 1
        return jnp.ones(self.n)

    def expected_objective_value(self):
        # The minimum objective value is 0.0
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: This implementation requires human review and verification against
# another CUTEst interface
class COOLHANSLS(AbstractUnconstrainedMinimisation):
    """A problem arising from the analysis of a Cooley-Hansen economy with
    loglinear approximation.

    The problem is to solve the matrix equation
               A * X * X + B * X + C = 0
    where A, B and C are known N times N matrices and X an unknown matrix
    of matching dimension. The instance considered here has N = 3.

    Source:
    S. Ceria, private communication, 1995.

    SIF input: Ph. Toint, Feb 1995.
    Least-squares version of COOLHANS.SIF, Nick Gould, Jan 2020.

    Classification: SUR2-RN-9-0
    """

    n: int = 9  # Number of variables (3x3 matrix X flattened)

    def objective(self, y, args):
        del args

        # Reshape the flattened variables into a 3x3 matrix
        X = y.reshape(3, 3)

        # Define matrices A, B, and C from the SIF file
        A = jnp.array([[0.0, 0.0, 0.0], [0.13725e-6, 937.62, -42.207], [0.0, 0.0, 0.0]])

        B = jnp.array(
            [
                [0.0060893, -44.292, 2.0011],
                [0.13880e-6, -1886.0, 42.362],
                [-0.13877e-6, 42.362, -2.0705],
            ]
        )

        C = jnp.array([[0.0, 44.792, 0.0], [0.0, 948.21, 0.0], [0.0, -42.684, 0.0]])

        # Compute A * X * X
        AXX = A @ X @ X

        # Compute B * X
        BX = B @ X

        # Matrix equation: A * X * X + B * X + C = 0
        # Residual matrix: A * X * X + B * X + C
        residual_matrix = AXX + BX + C

        # Compute the sum of squares of all elements in the residual matrix
        return jnp.sum(residual_matrix**2)

    def y0(self):
        # Initial values not explicitly specified in the SIF file,
        # but lines 95-98 suggest X(1,:) and X(3,:) might be fixed at 0
        # Default initialization is 0.0 for all variables
        return jnp.zeros(self.n)

    def args(self):
        return None

    def expected_result(self):
        # The SIF file doesn't specify the optimal solution
        # But line 146 suggests the objective value at the solution is 0.0,
        # meaning that a perfect solution to the matrix equation exists
        return None

    def expected_objective_value(self):
        # According to the SIF file, line 146
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HS55(AbstractConstrainedMinimisation):
    """Problem 55 from the Hock-Schittkowski test collection.

    A 6-variable nonlinear objective function with six linear equality constraints
    and bounds on variables.

    f(x) = x₁ + 2x₂ + 4x₅ + exp(x₁x₄)

    Subject to:
        x₁ + 2x₂ + 5x₅ - 6 = 0
        x₁ + x₂ + x₃ - 3 = 0
        x₄ + x₅ + x₆ - 2 = 0
        x₁ + x₄ - 1 = 0
        x₂ + x₅ - 2 = 0
        x₃ + x₆ - 2 = 0
        0 ≤ x₁ ≤ 1
        0 ≤ x₂
        0 ≤ x₃
        0 ≤ x₄ ≤ 1
        0 ≤ x₅
        0 ≤ x₆

    Source: problem 55 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    Also cited: Hsia [33]

    Classification: OLR2-AN-6-6
    """

    def objective(self, y, args):
        x1, x2, x3, x4, x5, x6 = y
        return x1 + 2 * x2 + 4 * x5 + jnp.exp(x1 * x4)

    def y0(self):
        return jnp.array([1.0, 2.0, 0.0, 0.0, 0.0, 2.0])

    def args(self):
        return None

    def expected_result(self):
        return jnp.array([0.0, 4.0 / 3.0, 5.0 / 3.0, 1.0, 2.0 / 3.0, 1.0 / 3.0])

    def expected_objective_value(self):
        return jnp.array(19.0 / 3.0)

    def bounds(self):
        lower = jnp.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        # From SIF file: only X1 and X4 have upper bounds of 1.0
        upper = jnp.array([1.0, jnp.inf, jnp.inf, 1.0, jnp.inf, jnp.inf])
        return lower, upper

    def constraint(self, y):
        x1, x2, x3, x4, x5, x6 = y
        # Equality constraints
        eq1 = x1 + 2 * x2 + 5 * x5 - 6
        eq2 = x1 + x2 + x3 - 3
        eq3 = x4 + x5 + x6 - 2
        eq4 = x1 + x4 - 1
        eq5 = x2 + x5 - 2
        eq6 = x3 + x6 - 2
        equality_constraints = jnp.array([eq1, eq2, eq3, eq4, eq5, eq6])
        return equality_constraints, None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HS27(AbstractConstrainedMinimisation):
    """Problem 27 from the Hock-Schittkowski test collection.

    A 3-variable quadratic function with one equality constraint.

    f(x) = 0.01(x₁ - 1)² + (x₂ - x₁²)²

    Subject to: x₁ + x₃² + 1 = 0

    Source: problem 27 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    Also cited: Miele e.al. [44,45]

    Classification: PQR-T1-6
    """

    def objective(self, y, args):
        x1, x2, x3 = y
        return 0.01 * (x1 - 1.0) ** 2 + (x2 - x1**2) ** 2

    def y0(self):
        return jnp.array([2.0, 2.0, 2.0])

    def args(self):
        return None

    def expected_result(self):
        return jnp.array([-1.0, 1.0, 0.0])

    def expected_objective_value(self):
        return jnp.array(0.04)

    def bounds(self):
        return None

    def constraint(self, y):
        x1, x2, x3 = y
        # Equality constraint: x₁ + x₃² + 1 = 0
        equality_constraint = x1 + x3**2 + 1.0
        return equality_constraint, None



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: This implementation requires human review and verification against
# another CUTEst interface
class BOXBODLS(AbstractUnconstrainedMinimisation):
    """NIST Data fitting problem BOXBOD.

    Fit: y = b1*(1-exp[-b2*x]) + e

    Source: Problem from the NIST nonlinear regression test set
    http://www.itl.nist.gov/div898/strd/nls/nls_main.shtml

    Reference: Box, G. P., W. G. Hunter, and J. S. Hunter (1978).
    Statistics for Experimenters, New York, NY: Wiley, pp. 483-487.

    SIF input: Nick Gould and Tyrone Rees, Oct 2015

    Classification: SUR2-MN-2-0
    """

    n: int = 2  # Problem has 2 variables
    m: int = 6  # Number of data points

    def objective(self, y, args):
        del args
        b1, b2 = y

        # Data points from the SIF file
        x_data = jnp.array([1.0, 2.0, 3.0, 5.0, 7.0, 10.0])
        y_data = jnp.array([109.0, 149.0, 149.0, 191.0, 213.0, 224.0])

        # Model: y = b1*(1-exp[-b2*x])
        y_pred = b1 * (1.0 - jnp.exp(-b2 * x_data))

        # Compute the residuals
        residuals = y_pred - y_data

        # Sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Initial values from SIF file (START1)
        return jnp.array([1.0, 1.0])

    def args(self):
        return None

    def expected_result(self):
        # NIST certified values from https://www.itl.nist.gov/div898/strd/nls/data/boxbod.shtml
        return jnp.array([213.80937, 0.54723])

    def expected_objective_value(self):
        # From NIST: sum of squared residuals is 1.1680088766E+03
        return jnp.array(1.1680088766e03)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class BT3(AbstractConstrainedMinimisation):
    """BT3 - Boggs-Tolle test problem 3.

    n = 5, m = 3.
    f(x) = (x₁ - x₂)² + (x₂ + x₃ - 2)² + (x₄ - 1)² + (x₅ - 1)².
    g₁(x) = x₁ + 3x₂.
    g₂(x) = x₃ + x₄ - 2x₅.
    g₃(x) = x₂ - x₅.

    Start 1: xᵢ = 2, i = 1, ..., 5.
    Start 2: xᵢ = 20, i = 1, ..., 5.
    Start 3: xᵢ = 200, i = 1, ..., 5.
    Solution: x* = (-0.76744, 0.25581, 0.62791, -0.11628, 0.25581).

    Source: Boggs, P.T. and Tolle, J.W.,
    "A strategy for global convergence in a sequential
    quadratic programming algorithm",
    SIAM J. Numer. Anal. 26(3), pp. 600-623, 1989.

    Classification: SLR2-AY-5-3
    """

    y0_id: int = 1  # Default to second starting point (20.0) as per SIF
    provided_y0s: frozenset = frozenset({0, 1, 2})

    def objective(self, y, args):
        del args
        x1, x2, x3, x4, x5 = y
        return (x1 - x2) ** 2 + (x2 + x3 - 2) ** 2 + (x4 - 1) ** 2 + (x5 - 1) ** 2

    def y0(self):
        if self.y0_id == 0:
            return jnp.array([2.0, 2.0, 2.0, 2.0, 2.0])
        elif self.y0_id == 1:
            return jnp.array([20.0, 20.0, 20.0, 20.0, 20.0])
        elif self.y0_id == 2:
            return jnp.array([200.0, 200.0, 200.0, 200.0, 200.0])

    def args(self):
        return None

    def expected_result(self):
        return jnp.array([-0.76744, 0.25581, 0.62791, -0.11628, 0.25581])

    def expected_objective_value(self):
        return None  # Not explicitly given

    def bounds(self):
        return None

    def constraint(self, y):
        x1, x2, x3, x4, x5 = y
        # Equality constraints
        g1 = x1 + 3 * x2
        g2 = x3 + x4 - 2 * x5
        g3 = x2 - x5
        equality_constraints = jnp.array([g1, g2, g3])
        return equality_constraints, None



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class DEVGLA2(AbstractUnconstrainedMinimisation):
    """DeVilliers-Glasser problem 2 from the SCIPY global optimization benchmark.

    This problem involves fitting a model of the form:
    y = x₁ * x₂^t * tanh(t * x₃ + sin(t * x₄)) * cos(t * e^x₅) + e

    to a set of data points (t, y) where t ranges from 0.1 to 1.6 in steps of 0.1
    and the y values are pre-computed based on specific parameter values.

    Source: Problem from the SCIPY benchmark set
    https://github.com/scipy/scipy/tree/master/benchmarks/benchmarks/go_benchmark_functions

    SIF input: Nick Gould, Jan 2020

    Classification: SUR2-MN-5-0
    """

    n: int = 5  # Number of variables

    def objective(self, y, args):
        del args
        x1, x2, x3, x4, x5 = y

        # Generate time points t from 0.0 to 1.5 with step 0.1 (16 points total)
        t_values = jnp.arange(0, 16) * 0.1

        # Compute the true data points (computed from SIF file with exact parameters)
        true_values = jnp.array(
            [
                0.000000,
                25.652980,
                40.985849,
                45.968797,
                44.800863,
                40.224285,
                33.489336,
                25.171629,
                15.612834,
                5.094175,
                -6.103169,
                -17.680274,
                -29.318700,
                -40.685106,
                -51.438706,
                -61.240116,
            ]
        )

        # Compute the model predictions
        model_pred = self._model(x1, x2, x3, x4, x5, t_values)

        # Calculate residuals
        residuals = model_pred - true_values

        # Sum of squared residuals
        return jnp.sum(residuals**2)

    def _model(self, x1, x2, x3, x4, x5, t_values):
        """Model function: x₁ * x₂^t * tanh(t * x₃ + sin(t * x₄)) * cos(t * e^x₅)"""
        # Compute x₂^t for all t values
        x2_power_t = x2**t_values

        # Compute sin(t * x₄) for all t values
        sin_tx4 = jnp.sin(t_values * x4)

        # Compute tanh(t * x₃ + sin(t * x₄)) for all t values
        tanh_term = jnp.tanh(t_values * x3 + sin_tx4)

        # Compute cos(t * e^x₅) for all t values
        cos_term = jnp.cos(t_values * jnp.exp(x5))

        # Compute the final model prediction
        return x1 * x2_power_t * tanh_term * cos_term

    def y0(self):
        # Initial values from SIF file
        return jnp.array([20.0, 2.0, 2.0, 2.0, 0.2])

    def args(self):
        return None

    def expected_result(self):
        # Solution values given in the SIF file
        return jnp.array([53.81, 1.27, 3.012, 2.13, 0.507])

    def expected_objective_value(self):
        # For a perfect fit, the objective value would be 0
        return jnp.array(0.0)



import jax
import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: needs human review
class ENSOLS(AbstractUnconstrainedMinimisation):
    """ENSOLS - El Niño-Southern Oscillation data fitting.

    This problem involves fitting a sum of sinusoids with different periods
    to the El Niño-Southern Oscillation (ENSO) time series data, which measures
    the monthly averaged atmospheric pressure differences between Easter Island
    and Darwin, Australia.

    The model function is:
        y = b1 + b2*cos(2πt/12) + b3*sin(2πt/12) +
            b5*cos(2πt/b4) + b6*sin(2πt/b4) +
            b8*cos(2πt/b7) + b9*sin(2πt/b7)

    Source: NIST nonlinear regression test set
    http://www.itl.nist.gov/div898/strd/nls/nls_main.shtml

    Reference:
    Kahaner, D., C. Moler, and S. Nash, (1989).
    Numerical Methods and Software.
    Englewood Cliffs, NJ: Prentice Hall, pp. 441-445.

    SIF input: Ph. Toint, Apr 1997.

    Classification: SUR2-MN-9-0
    """

    # Allow selecting which starting point to use (0-based indexing)
    start_point: int = 0  # 0 or 1
    _allowed_start_points = frozenset({0, 1})

    def __check_init__(self):
        if self.start_point not in self._allowed_start_points:
            allowed = self._allowed_start_points
            msg = f"start_point must be in {allowed}, got {self.start_point}"
            raise ValueError(msg)

    def objective(self, y, args):
        del args

        # Extract the 9 parameters
        b1, b2, b3, b4, b5, b6, b7, b8, b9 = y

        # Time points: 1 to 168 months
        t = jnp.arange(1, 169)

        # ENSO observed data from the SIF file
        enso_data = jnp.array(
            [
                12.9,
                11.3,
                10.6,
                11.2,
                10.9,
                7.5,
                7.7,
                11.7,
                12.9,
                14.3,
                10.9,
                13.7,
                17.1,
                14.0,
                15.3,
                8.5,
                5.7,
                5.5,
                7.6,
                8.6,
                7.3,
                7.6,
                12.7,
                11.0,
                12.7,
                12.9,
                13.0,
                10.9,
                10.4,
                10.2,
                8.0,
                10.9,
                13.6,
                10.5,
                9.2,
                12.4,
                12.7,
                13.3,
                10.1,
                7.8,
                4.8,
                3.0,
                2.5,
                6.3,
                9.7,
                11.6,
                8.6,
                12.4,
                10.5,
                13.3,
                10.4,
                8.1,
                3.7,
                1.0,
                1.2,
                2.3,
                4.5,
                6.6,
                5.0,
                10.3,
                9.2,
                12.1,
                10.3,
                9.5,
                5.3,
                3.8,
                2.0,
                0.8,
                0.2,
                0.9,
                2.0,
                6.7,
                7.9,
                11.2,
                8.4,
                7.1,
                5.6,
                3.5,
                1.0,
                1.2,
                1.5,
                5.1,
                3.3,
                9.8,
                9.1,
                11.9,
                10.4,
                9.5,
                4.4,
                2.9,
                -0.3,
                1.4,
                3.5,
                5.3,
                8.2,
                11.0,
                11.8,
                11.8,
                8.3,
                9.8,
                5.3,
                3.9,
                1.9,
                3.7,
                3.5,
                5.3,
                5.9,
                9.1,
                9.4,
                12.2,
                9.5,
                7.1,
                4.8,
                2.9,
                1.2,
                0.1,
                -0.8,
                1.9,
                0.3,
                6.6,
                9.9,
                12.5,
                9.9,
                6.4,
                5.1,
                3.6,
                1.8,
                0.7,
                0.4,
                1.4,
                2.5,
                5.5,
                8.7,
                14.3,
                10.0,
                6.5,
                3.0,
                2.4,
                2.0,
                0.8,
                0.9,
                5.1,
                5.5,
                8.0,
                10.3,
                13.5,
                9.8,
                5.2,
                1.8,
                0.5,
                1.3,
                3.9,
                5.0,
                5.9,
                5.6,
                6.9,
                10.1,
                11.5,
                9.5,
                6.4,
                5.1,
                2.4,
                2.0,
                4.0,
                3.8,
                4.9,
                6.8,
                9.7,
            ]
        )

        # Calculate the model values at each time point
        def model_func(t_i):
            # Annual cycle terms (fixed period of 12 months)
            annual_cos = b2 * jnp.cos(2.0 * jnp.pi * t_i / 12.0)
            annual_sin = b3 * jnp.sin(2.0 * jnp.pi * t_i / 12.0)

            # First variable period terms
            period1_cos = b5 * jnp.cos(2.0 * jnp.pi * t_i / b4)
            period1_sin = b6 * jnp.sin(2.0 * jnp.pi * t_i / b4)

            # Second variable period terms
            period2_cos = b8 * jnp.cos(2.0 * jnp.pi * t_i / b7)
            period2_sin = b9 * jnp.sin(2.0 * jnp.pi * t_i / b7)

            # Combine all terms
            return (
                b1
                + annual_cos
                + annual_sin
                + period1_cos
                + period1_sin
                + period2_cos
                + period2_sin
            )

        # Calculate model predictions for all time points at once
        predictions = jax.vmap(model_func)(t)

        # Compute residuals
        residuals = predictions - enso_data

        # Return the sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Two starting points from the SIF file
        if self.start_point == 0:
            # Starting point 1
            return jnp.array([11.0, 3.0, 0.5, 40.0, -0.7, -1.3, 25.0, -0.3, 1.4])
        else:  # self.start_point == 1
            # Starting point 2
            return jnp.array([10.0, 3.0, 0.5, 44.0, -1.5, 0.5, 26.0, -0.1, 1.5])

    def args(self):
        return None

    def expected_result(self):
        # Should use certified values from NIST
        return None

    def expected_objective_value(self):
        # Should use certified minimum value from NIST
        return None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HS93(AbstractConstrainedMinimisation):
    """Problem 93 from the Hock-Schittkowski test collection.

    A 6-variable nonlinear objective function with two inequality constraints.

    f(x) = 0.0204*x₁*x₄*(x₁ + x₂ + x₃) + 0.0187*x₂*x₃*(x₁ + 1.57*x₂ + x₄)
           + 0.0607*x₁*x₄*x₅²*(x₁ + x₂ + x₃) + 0.0437*x₂*x₃*x₆²*(x₁ + 1.57*x₂ + x₄)

    Subject to:
        0.001*x₁*x₂*x₃*x₄*x₅*x₆ ≥ 2.07
        0.00062*x₁*x₄*x₅²*(x₁ + x₂ + x₃) + 0.00058*x₂*x₃*x₆²*(x₁ + 1.57*x₂ + x₄) ≤ 1
        0 ≤ xᵢ, i=1,...,6

    Source: problem 93 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    Classification: PGR-P1-5
    """

    def objective(self, y, args):
        x1, x2, x3, x4, x5, x6 = y
        return (
            0.0204 * x1 * x4 * (x1 + x2 + x3)
            + 0.0187 * x2 * x3 * (x1 + 1.57 * x2 + x4)
            + 0.0607 * x1 * x4 * x5**2 * (x1 + x2 + x3)
            + 0.0437 * x2 * x3 * x6**2 * (x1 + 1.57 * x2 + x4)
        )

    def y0(self):
        return jnp.array([5.54, 4.4, 12.02, 11.82, 0.702, 0.852])

    def args(self):
        return None

    def expected_result(self):
        return jnp.array([5.332666, 4.656744, 10.43299, 12.08230, 0.7526074, 0.8420251])

    def expected_objective_value(self):
        return jnp.array(135.075961)

    def bounds(self):
        return [(0.0, None) for _ in range(6)]

    def constraint(self, y):
        x1, x2, x3, x4, x5, x6 = y
        # Inequality constraints from SIF file
        # C1 is a 'G' type: 0.001 * x1*x2*x3*x4*x5*x6 >= 2.07
        # pycutest reports the raw constraint value
        ineq1 = 0.001 * x1 * x2 * x3 * x4 * x5 * x6 - 2.07

        # C2 is an 'L' type constraint
        # 0.00062*OE3 + 0.00058*OE4 <= 1
        # where OE3 = x1*x4*x5²*(x1 + x2 + x3)
        # and OE4 = x2*x3*x6²*(x1 + 1.57*x2 + x4)
        oe3 = x1 * x4 * x5**2 * (x1 + x2 + x3)
        oe4 = x2 * x3 * x6**2 * (x1 + 1.57 * x2 + x4)
        ineq2 = 0.00062 * oe3 + 0.00058 * oe4 - 1.0

        inequality_constraints = jnp.array([ineq1, ineq2])
        return None, inequality_constraints



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class BT6(AbstractConstrainedMinimisation):
    """BT6 - Boggs-Tolle test problem 6.

    n = 5, m = 2.
    f(x) = (x₁ - 1)² + (x₁ - x₂)² + (x₃ - 1)² + (x₄ - 1)⁴ + (x₅ - 1)⁶.
    g₁(x) = x₄x₁² + sin(x₄ - x₅) - 2√2.
    g₂(x) = x₂ + x₃⁴x₅² - 8 - √2.

    Start 1: xᵢ = 2, i = 1, ..., 5.
    Start 2: xᵢ = 8, i = 1, ..., 5.
    Solution: x* = (1.1662, 1.1821, 1.3803, 1.5060, 0.61092).

    Source: Boggs, P.T. and Tolle, J.W.,
    "A strategy for global convergence in a sequential
    quadratic programming algorithm",
    SIAM J. Numer. Anal. 26(3), pp. 600-623, 1989.

    Classification: OOR2-AY-5-2
    """

    y0_id: int = 0
    provided_y0s: frozenset = frozenset({0, 1})

    def objective(self, y, args):
        del args
        x1, x2, x3, x4, x5 = y
        return (
            (x1 - 1) ** 2
            + (x1 - x2) ** 2
            + (x3 - 1) ** 2
            + (x4 - 1) ** 4
            + (x5 - 1) ** 6
        )

    def y0(self):
        if self.y0_id == 0:
            return jnp.array([2.0, 2.0, 2.0, 2.0, 2.0])
        elif self.y0_id == 1:
            return jnp.array([8.0, 8.0, 8.0, 8.0, 8.0])

    def args(self):
        return None

    def expected_result(self):
        return jnp.array([1.1662, 1.1821, 1.3803, 1.5060, 0.61092])

    def expected_objective_value(self):
        return None  # Not explicitly given

    def bounds(self):
        return None

    def constraint(self, y):
        x1, x2, x3, x4, x5 = y
        # Equality constraints
        g1 = x4 * x1**2 + jnp.sin(x4 - x5) - 2 * jnp.sqrt(2)
        g2 = x2 + x3**4 * x5**2 - 8 - jnp.sqrt(2)
        equality_constraints = jnp.array([g1, g2])
        return equality_constraints, None



from ._bounded_minimisation import (
    bounded_minimisation_problems as bounded_minimisation_problems,
    EXP2B as EXP2B,
    HS1 as HS1,
    HS2 as HS2,
    HS3 as HS3,
    HS4 as HS4,
    HS5 as HS5,
    HS25 as HS25,
    HS38 as HS38,
    HS45 as HS45,
    HS110 as HS110,
)
from ._constrained_minimisation import (
    BT1 as BT1,
    BT2 as BT2,
    BT3 as BT3,
    BT4 as BT4,
    BT5 as BT5,
    BT6 as BT6,
    BT7 as BT7,
    BT8 as BT8,
    BT9 as BT9,
    BT10 as BT10,
    BT11 as BT11,
    BT12 as BT12,
    BT13 as BT13,
    constrained_minimisation_problems as constrained_minimisation_problems,
    HS6 as HS6,
    HS7 as HS7,
    HS8 as HS8,
    HS9 as HS9,
    HS10 as HS10,
    HS11 as HS11,
    HS12 as HS12,
    HS13 as HS13,
    HS14 as HS14,
    HS15 as HS15,
    HS16 as HS16,
    HS17 as HS17,
    HS18 as HS18,
    HS19 as HS19,
    HS20 as HS20,
    HS21 as HS21,
    HS22 as HS22,
    HS23 as HS23,
    HS24 as HS24,
    HS26 as HS26,
    HS27 as HS27,
    HS28 as HS28,
    HS29 as HS29,
    HS30 as HS30,
    HS31 as HS31,
    HS32 as HS32,
    HS33 as HS33,
    HS34 as HS34,
    HS35 as HS35,
    HS36 as HS36,
    HS37 as HS37,
    HS39 as HS39,
    HS40 as HS40,
    HS41 as HS41,
    HS42 as HS42,
    HS43 as HS43,
    HS44 as HS44,
    HS46 as HS46,
    HS47 as HS47,
    HS48 as HS48,
    HS49 as HS49,
    HS50 as HS50,
    HS51 as HS51,
    HS52 as HS52,
    HS53 as HS53,
    HS54 as HS54,
    HS55 as HS55,
    HS56 as HS56,
    HS57 as HS57,
    # HS59 as HS59,  # TODO: Human review - objective function discrepancy
    HS60 as HS60,
    HS61 as HS61,
    HS62 as HS62,
    HS63 as HS63,
    HS64 as HS64,
    HS65 as HS65,
    HS66 as HS66,
    # HS67 as HS67,  # TODO: Human review - different SIF file version
    HS68 as HS68,
    HS69 as HS69,
    HS71 as HS71,
    HS72 as HS72,
    HS73 as HS73,
    HS74 as HS74,
    HS75 as HS75,
    HS76 as HS76,
    HS77 as HS77,
    HS78 as HS78,
    HS79 as HS79,
    HS80 as HS80,
    HS81 as HS81,
    HS83 as HS83,
    HS93 as HS93,
    # HS99 as HS99,  # TODO: Needs human review - complex recursive formulation
    HS100 as HS100,
    HS101 as HS101,
    HS102 as HS102,
    HS103 as HS103,
    HS104 as HS104,
    HS105 as HS105,
    HS106 as HS106,
    HS107 as HS107,
    HS108 as HS108,
    HS109 as HS109,
    HS111 as HS111,
    HS112 as HS112,
    HS113 as HS113,
    HS114 as HS114,
    HS116 as HS116,
    HS117 as HS117,
    HS118 as HS118,
    HS119 as HS119,
    LUKVLE1 as LUKVLE1,
    # LUKVLE2 as LUKVLE2,
    LUKVLE3 as LUKVLE3,
    # LUKVLE4 as LUKVLE4,  # Use LUKVLE4C instead
    # LUKVLE4C as LUKVLE4C,
    LUKVLE5 as LUKVLE5,
    LUKVLE6 as LUKVLE6,
    LUKVLE7 as LUKVLE7,
    LUKVLE8 as LUKVLE8,
    LUKVLE9 as LUKVLE9,
    LUKVLE10 as LUKVLE10,
    LUKVLE11 as LUKVLE11,
    # LUKVLE12 as LUKVLE12,  # Has constraint function inconsistencies
    LUKVLE13 as LUKVLE13,
    LUKVLE14 as LUKVLE14,
    LUKVLE15 as LUKVLE15,
    LUKVLE16 as LUKVLE16,
    LUKVLE17 as LUKVLE17,
    LUKVLE18 as LUKVLE18,
    LUKVLI1 as LUKVLI1,
    # LUKVLI2 as LUKVLI2,
    LUKVLI3 as LUKVLI3,
    # LUKVLI4 as LUKVLI4,  # Use LUKVLI4C instead
    # LUKVLI4C as LUKVLI4C,
    LUKVLI5 as LUKVLI5,
    LUKVLI6 as LUKVLI6,
    LUKVLI7 as LUKVLI7,
    LUKVLI8 as LUKVLI8,
    LUKVLI9 as LUKVLI9,
    LUKVLI10 as LUKVLI10,
    LUKVLI11 as LUKVLI11,
    # LUKVLI12 as LUKVLI12,  # Has constraint function inconsistencies
    LUKVLI13 as LUKVLI13,
    LUKVLI14 as LUKVLI14,
    LUKVLI15 as LUKVLI15,
    LUKVLI16 as LUKVLI16,
    LUKVLI17 as LUKVLI17,
    LUKVLI18 as LUKVLI18,
)
from ._unconstrained_minimisation import (
    AKIVA as AKIVA,
    ALLINITU as ALLINITU,
    ARGLINA as ARGLINA,
    ARGLINB as ARGLINB,
    ARGLINC as ARGLINC,
    ARGTRIGLS as ARGTRIGLS,
    ARWHEAD as ARWHEAD,
    # BA_L1LS as BA_L1LS,  # TODO: BA_L family needs to be split into separate files
    # BA_L1SPLS as BA_L1SPLS,  # TODO: BA_L family needs to be split into separate files
    BARD as BARD,
    BDQRTIC as BDQRTIC,
    BEALE as BEALE,
    BIGGS6 as BIGGS6,
    BOX as BOX,
    BOX3 as BOX3,
    BOXBODLS as BOXBODLS,
    # BOXPOWER as BOXPOWER,  # TODO: Human review - minor gradient discrepancy
    # BRKMCC as BRKMCC,  # TODO: Human review - significant discrepancies
    # BROWNAL as BROWNAL,  # TODO: Human review - small Hessian discrepancies
    BROWNBS as BROWNBS,
    BROWNDEN as BROWNDEN,
    BROYDN3DLS as BROYDN3DLS,
    BROYDN7D as BROYDN7D,
    # BROYDNBDLS as BROYDNBDLS,  # TODO: Gradient test fails - needs human review
    # BRYBND as BRYBND,  # TODO: Gradient test fails - needs human review
    # CERI651ALS as CERI651ALS,  # TODO: Human review - numerical instability
    # CERI651BLS as CERI651BLS,  # TODO: Human review - numerical instability
    # CERI651CLS as CERI651CLS,  # TODO: Human review - numerical instability
    # CERI651DLS as CERI651DLS,  # TODO: Human review - numerical instability
    # CERI651ELS as CERI651ELS,  # TODO: Human review - numerical instability
    CHAINWOO as CHAINWOO,
    CHNROSNB as CHNROSNB,
    CHNRSNBM as CHNRSNBM,
    CHWIRUT1LS as CHWIRUT1LS,
    CHWIRUT2LS as CHWIRUT2LS,
    CLIFF as CLIFF,
    CLUSTERLS as CLUSTERLS,
    COATING as COATING,
    COOLHANSLS as COOLHANSLS,
    COSINE as COSINE,
    CRAGGLVY as CRAGGLVY,
    CUBE as CUBE,
    CURLY10 as CURLY10,
    CURLY20 as CURLY20,
    CURLY30 as CURLY30,
    # CYCLOOCFLS as CYCLOOCFLS,  # TODO: Human review - times out with default p=10000
    DANIWOODLS as DANIWOODLS,
    DENSCHNA as DENSCHNA,
    DENSCHNB as DENSCHNB,
    DENSCHNC as DENSCHNC,
    DENSCHND as DENSCHND,
    DENSCHNE as DENSCHNE,
    DENSCHNF as DENSCHNF,
    DEVGLA1 as DEVGLA1,
    DEVGLA2 as DEVGLA2,
    DIXMAANA1 as DIXMAANA1,
    DIXMAANB as DIXMAANB,
    DIXMAANC as DIXMAANC,
    DIXMAAND as DIXMAAND,
    DIXMAANE1 as DIXMAANE1,
    DIXMAANF as DIXMAANF,
    DIXMAANG as DIXMAANG,
    DIXMAANH as DIXMAANH,
    DIXMAANI1 as DIXMAANI1,
    DIXMAANJ as DIXMAANJ,
    DIXMAANK as DIXMAANK,
    DIXMAANL as DIXMAANL,
    DIXMAANM1 as DIXMAANM1,
    DIXMAANN as DIXMAANN,
    DIXMAANO as DIXMAANO,
    DIXMAANP as DIXMAANP,
    DIXON3DQ as DIXON3DQ,
    DJTL as DJTL,
    DQDRTIC as DQDRTIC,
    DQRTIC as DQRTIC,
    # ECKERLE4LS as ECKERLE4LS,  # TODO: Human review - significant discrepancies
    EDENSCH as EDENSCH,
    EG2 as EG2,
    EGGCRATE as EGGCRATE,
    EIGENALS as EIGENALS,
    EIGENBLS as EIGENBLS,
    EIGENCLS as EIGENCLS,
    ELATVIDU as ELATVIDU,
    ENGVAL1 as ENGVAL1,
    ENGVAL2 as ENGVAL2,
    # ENSOLS as ENSOLS,  # TODO: Human review - significant discrepancies
    ERRINROS as ERRINROS,
    # ERRINRSM as ERRINRSM,  # TODO: Human review - significant discrepancies
    EXP2 as EXP2,
    EXPFIT as EXPFIT,
    # EXTROSNB as EXTROSNB,  # TODO: Human review - objective/gradient discrepancies
    # FBRAIN3LS as FBRAIN3LS,  # TODO: Human review - complex data dependencies
    FLETBV3M as FLETBV3M,
    FLETCBV2 as FLETCBV2,
    FLETCBV3 as FLETCBV3,
    # FLETCHBV as FLETCHBV,  # TODO: Human review - objective/gradient discrepancies
    FLETCHCR as FLETCHCR,
    # FMINSRF2 as FMINSRF2,  # TODO: Human review - starting value/gradient issues
    # FMINSURF as FMINSURF,  # TODO: Human review - starting value/gradient issues
    # FREURONE as FREURONE,  # TODO: Human review - miscategorized (constrained)
    FREUROTH as FREUROTH,
    # GAUSS1LS as GAUSS1LS,  # TODO: Human review - issues reported by user
    # GAUSS2LS as GAUSS2LS,  # TODO: Human review - issues reported by user
    # GAUSS3LS as GAUSS3LS,  # TODO: Human review - issues reported by user
    GAUSSIAN as GAUSSIAN,
    # GBRAINLS as GBRAINLS,  # TODO: Human review - complex data dependencies
    GENHUMPS as GENHUMPS,
    GENROSE as GENROSE,
    GROWTHLS as GROWTHLS,
    # GULF as GULF,  # TODO: Human review - issues reported by user
    HAHN1LS as HAHN1LS,
    HAIRY as HAIRY,
    # HATFLDD as HATFLDD,  # TODO: HATFLD family needs human review - discrepancies
    # HATFLDE as HATFLDE,  # TODO: HATFLD family needs human review - discrepancies
    # HATFLDFL as HATFLDFL,  # TODO: HATFLD family needs human review - discrepancies
    # HATFLDFLS as HATFLDFLS,  # TODO: HATFLD family needs human review - discrepancies
    # HATFLDGLS as HATFLDGLS,  # TODO: HATFLD family needs human review - discrepancies
    # HEART6LS as HEART6LS,  # TODO: Human review - significant discrepancies
    # HEART8LS as HEART8LS,  # TODO: Human review - significant discrepancies
    HELIX as HELIX,
    # HIELOW as HIELOW,  # TODO: Human review - significant discrepancies
    HILBERTA as HILBERTA,
    HILBERTB as HILBERTB,
    # HIMMELBB as HIMMELBB,  # TODO: Human review - Hessian discrepancies
    HIMMELBCLS as HIMMELBCLS,
    # HIMMELBF as HIMMELBF,  # TODO: Human review - Hessian discrepancies
    HIMMELBG as HIMMELBG,
    HIMMELBH as HIMMELBH,
    HUMPS as HUMPS,
    INDEF as INDEF,
    INTEQNELS as INTEQNELS,
    JENSMP as JENSMP,
    JUDGE as JUDGE,
    KIRBY2LS as KIRBY2LS,
    KOWOSB as KOWOSB,
    # KSSLS as KSSLS,  # TODO: Human review - significant obj/grad discrepancies
    LANCZOS1LS as LANCZOS1LS,
    LANCZOS2LS as LANCZOS2LS,
    LIARWHD as LIARWHD,
    LSC1LS as LSC1LS,
    LSC2LS as LSC2LS,
    # MANCINO as MANCINO,  # TODO: Human review - significant discrepancies in all
    ROSENBR as ROSENBR,
    TENFOLDTRLS as TENFOLDTRLS,
    unconstrained_minimisation_problems as unconstrained_minimisation_problems,
)


problems_dict = {
    "HS1": HS1(),
    "HS2": HS2(),
    "HS3": HS3(),
    "HS4": HS4(),
    "HS5": HS5(),
    "HS6": HS6(),
    "HS7": HS7(),
    "HS8": HS8(),
    "HS9": HS9(),
    "HS10": HS10(),
    "HS11": HS11(),
    "HS12": HS12(),
    "HS13": HS13(),
    "HS14": HS14(),
    "HS15": HS15(),
    "HS16": HS16(),
    "HS17": HS17(),
    "HS18": HS18(),
    "HS19": HS19(),
    "HS20": HS20(),
    "HS21": HS21(),
    "HS22": HS22(),
    "HS23": HS23(),
    "HS24": HS24(),
    "HS25": HS25(),
    "HS26": HS26(),
    "HS27": HS27(),
    "HS28": HS28(),
    "HS29": HS29(),
    "HS30": HS30(),
    "HS31": HS31(),
    "HS32": HS32(),
    "HS33": HS33(),
    "HS34": HS34(),
    "HS35": HS35(),
    "HS36": HS36(),
    "HS37": HS37(),
    "HS38": HS38(),
    "HS39": HS39(),
    "HS40": HS40(),
    "HS41": HS41(),
    "HS42": HS42(),
    "HS43": HS43(),
    "HS44": HS44(),
    "HS45": HS45(),
    "HS46": HS46(),
    "HS47": HS47(),
    "HS48": HS48(),
    "HS49": HS49(),
    "HS50": HS50(),
    "HS51": HS51(),
    "HS52": HS52(),
    "HS53": HS53(),
    "HS54": HS54(),
    "HS55": HS55(),
    "HS56": HS56(),
    "HS57": HS57(),
    # "HS59": HS59(),  # TODO: Human review - objective function discrepancy
    "HS60": HS60(),
    "HS61": HS61(),
    "HS62": HS62(),
    "HS63": HS63(),
    "HS64": HS64(),
    "HS65": HS65(),
    "HS66": HS66(),
    # "HS67": HS67(),  # TODO: Human review - different SIF file version
    "HS68": HS68(),
    "HS69": HS69(),
    "HS71": HS71(),
    "HS72": HS72(),
    "HS73": HS73(),
    "HS74": HS74(),
    "HS75": HS75(),
    "HS76": HS76(),
    "HS77": HS77(),
    "HS78": HS78(),
    "HS79": HS79(),
    "HS80": HS80(),
    "HS81": HS81(),
    "HS83": HS83(),
    "HS93": HS93(),
    # "HS99": HS99(),  # TODO: Needs human review - complex recursive formulation
    "HS100": HS100(),
    "HS101": HS101(),
    "HS102": HS102(),
    "HS103": HS103(),
    "HS104": HS104(),
    "HS105": HS105(),
    "HS106": HS106(),
    "HS107": HS107(),
    "HS108": HS108(),
    "HS109": HS109(),
    "HS110": HS110(),
    "HS111": HS111(),
    "HS112": HS112(),
    "HS113": HS113(),
    "HS114": HS114(),
    "HS116": HS116(),
    "HS117": HS117(),
    "HS118": HS118(),
    "HS119": HS119(),
    "BT1": BT1(),
    "BT2": BT2(),
    "BT3": BT3(),
    "BT4": BT4(),
    "BT5": BT5(),
    "BT6": BT6(),
    "BT7": BT7(),
    "BT8": BT8(),
    "BT9": BT9(),
    "BT10": BT10(),
    "BT11": BT11(),
    "BT12": BT12(),
    "BT13": BT13(),
    "LUKVLE1": LUKVLE1(),
    # "LUKVLE2": LUKVLE2(),
    "LUKVLE3": LUKVLE3(),
    # "LUKVLE4": LUKVLE4(),  # Use LUKVLE4C instead
    # "LUKVLE4C": LUKVLE4C(),
    "LUKVLE5": LUKVLE5(),
    "LUKVLE6": LUKVLE6(),
    "LUKVLE7": LUKVLE7(),
    "LUKVLE8": LUKVLE8(),
    "LUKVLE9": LUKVLE9(),
    "LUKVLE10": LUKVLE10(),
    "LUKVLE11": LUKVLE11(),
    # "LUKVLE12": LUKVLE12(),  # Has constraint function inconsistencies
    "LUKVLE13": LUKVLE13(),
    "LUKVLE14": LUKVLE14(),
    "LUKVLE15": LUKVLE15(),
    "LUKVLE16": LUKVLE16(),
    "LUKVLE17": LUKVLE17(),
    "LUKVLE18": LUKVLE18(),
    "LUKVLI1": LUKVLI1(),
    # "LUKVLI2": LUKVLI2(),
    "LUKVLI3": LUKVLI3(),
    # "LUKVLI4": LUKVLI4(),  # Use LUKVLI4C instead
    # "LUKVLI4C": LUKVLI4C(),
    "LUKVLI5": LUKVLI5(),
    "LUKVLI6": LUKVLI6(),
    "LUKVLI7": LUKVLI7(),
    "LUKVLI8": LUKVLI8(),
    "LUKVLI9": LUKVLI9(),
    "LUKVLI10": LUKVLI10(),
    "LUKVLI11": LUKVLI11(),
    # "LUKVLI12": LUKVLI12(),  # Has constraint function inconsistencies
    "LUKVLI13": LUKVLI13(),
    "LUKVLI14": LUKVLI14(),
    "LUKVLI15": LUKVLI15(),
    "LUKVLI16": LUKVLI16(),
    "LUKVLI17": LUKVLI17(),
    "LUKVLI18": LUKVLI18(),
    "AKIVA": AKIVA(),
    "ALLINITU": ALLINITU(),
    "ARGLINA": ARGLINA(),
    "ARGLINB": ARGLINB(),
    "ARGLINC": ARGLINC(),
    "ARGTRIGLS": ARGTRIGLS(),
    "ARWHEAD": ARWHEAD(),
    # "BA_L1LS": BA_L1LS(),  # TODO: BA_L family needs to be split into files
    # "BA_L1SPLS": BA_L1SPLS(),  # TODO: BA_L family needs to be split into files
    "BARD": BARD(),
    "BDQRTIC": BDQRTIC(),
    "BEALE": BEALE(),
    "BIGGS6": BIGGS6(),
    "BOX": BOX(),
    "BOX3": BOX3(),
    "BOXBODLS": BOXBODLS(),
    # "BOXPOWER": BOXPOWER(),  # TODO: Human review - minor gradient discrepancy
    # "BRKMCC": BRKMCC(),  # TODO: Human review - significant discrepancies
    # "BROWNAL": BROWNAL(),  # TODO: Human review - small Hessian discrepancies
    "BROWNBS": BROWNBS(),
    "BROWNDEN": BROWNDEN(),
    "BROYDN3DLS": BROYDN3DLS(),
    "BROYDN7D": BROYDN7D(),
    # "BROYDNBDLS": BROYDNBDLS(),  # TODO: Gradient test fails - needs human review
    # "BRYBND": BRYBND(),  # TODO: Gradient test fails - needs human review
    # "CERI651ALS": CERI651ALS(),  # TODO: Human review - numerical instability
    # "CERI651BLS": CERI651BLS(),  # TODO: Human review - numerical instability
    # "CERI651CLS": CERI651CLS(),  # TODO: Human review - numerical instability
    # "CERI651DLS": CERI651DLS(),  # TODO: Human review - numerical instability
    # "CERI651ELS": CERI651ELS(),  # TODO: Human review - numerical instability
    "CHAINWOO": CHAINWOO(),
    "CHNROSNB": CHNROSNB(),
    "CHNRSNBM": CHNRSNBM(),
    "CHWIRUT1LS": CHWIRUT1LS(),
    "CHWIRUT2LS": CHWIRUT2LS(),
    "CLIFF": CLIFF(),
    "CLUSTERLS": CLUSTERLS(),
    "COATING": COATING(),
    "COOLHANSLS": COOLHANSLS(),
    "COSINE": COSINE(),
    "CRAGGLVY": CRAGGLVY(),
    "CUBE": CUBE(),
    "CURLY10": CURLY10(),
    "CURLY20": CURLY20(),
    "CURLY30": CURLY30(),
    # "CYCLOOCFLS": CYCLOOCFLS(),  # TODO: Human review - times out with default p=10000
    "DANIWOODLS": DANIWOODLS(),
    "DENSCHNA": DENSCHNA(),
    "DENSCHNB": DENSCHNB(),
    "DENSCHNC": DENSCHNC(),
    "DENSCHND": DENSCHND(),
    "DENSCHNE": DENSCHNE(),
    "DENSCHNF": DENSCHNF(),
    "DEVGLA1": DEVGLA1(),
    "DEVGLA2": DEVGLA2(),
    "DIXMAANA1": DIXMAANA1(),
    "DIXMAANB": DIXMAANB(),
    "DIXMAANC": DIXMAANC(),
    "DIXMAAND": DIXMAAND(),
    "DIXMAANE1": DIXMAANE1(),
    "DIXMAANF": DIXMAANF(),
    "DIXMAANG": DIXMAANG(),
    "DIXMAANH": DIXMAANH(),
    "DIXMAANI1": DIXMAANI1(),
    "DIXMAANJ": DIXMAANJ(),
    "DIXMAANK": DIXMAANK(),
    "DIXMAANL": DIXMAANL(),
    "DIXMAANM1": DIXMAANM1(),
    "DIXMAANN": DIXMAANN(),
    "DIXMAANO": DIXMAANO(),
    "DIXMAANP": DIXMAANP(),
    "DIXON3DQ": DIXON3DQ(),
    "DJTL": DJTL(),
    "DQDRTIC": DQDRTIC(),
    "DQRTIC": DQRTIC(),
    # "ECKERLE4LS": ECKERLE4LS(),  # TODO: Human review - significant discrepancies
    "EDENSCH": EDENSCH(),
    "EG2": EG2(),
    "EGGCRATE": EGGCRATE(),
    "EIGENALS": EIGENALS(),
    "EIGENBLS": EIGENBLS(),
    "EIGENCLS": EIGENCLS(),
    "ELATVIDU": ELATVIDU(),
    "ENGVAL1": ENGVAL1(),
    "ENGVAL2": ENGVAL2(),
    # "ENSOLS": ENSOLS(),  # TODO: Human review - significant discrepancies
    "ERRINROS": ERRINROS(),
    # "ERRINRSM": ERRINRSM(),  # TODO: Human review - significant discrepancies
    "EXP2": EXP2(),
    "EXP2B": EXP2B(),
    "EXPFIT": EXPFIT(),
    # "EXTROSNB": EXTROSNB(),  # TODO: Human review - objective/gradient discrepancies
    # "FBRAIN3LS": FBRAIN3LS(),  # TODO: Human review - complex data dependencies
    # "FLETCHBV": FLETCHBV(),  # TODO: Human review - objective/gradient discrepancies
    "FLETBV3M": FLETBV3M(),
    "FLETCBV2": FLETCBV2(),
    "FLETCHCR": FLETCHCR(),
    "FLETCBV3": FLETCBV3(),
    # "FMINSRF2": FMINSRF2(),  # TODO: Human review - starting value/gradient issues
    # "FMINSURF": FMINSURF(),  # TODO: Human review - starting value/gradient issues
    # "FREURONE": FREURONE(),  # TODO: Human review - miscategorized (constrained)
    "FREUROTH": FREUROTH(),
    # "GAUSS1LS": GAUSS1LS(),  # TODO: Human review - issues reported by user
    # "GAUSS2LS": GAUSS2LS(),  # TODO: Human review - issues reported by user
    # "GAUSS3LS": GAUSS3LS(),  # TODO: Human review - issues reported by user
    "GAUSSIAN": GAUSSIAN(),
    # "GBRAINLS": GBRAINLS(),  # TODO: Human review - complex data dependencies
    "GENHUMPS": GENHUMPS(),
    "GENROSE": GENROSE(),
    "GROWTHLS": GROWTHLS(),
    # "GULF": GULF(),  # TODO: Human review - issues reported by user
    "HAHN1LS": HAHN1LS(),
    "HAIRY": HAIRY(),
    # "HATFLDD": HATFLDD(),  # TODO: HATFLD family needs human review - discrepancies
    # "HATFLDE": HATFLDE(),  # TODO: HATFLD family needs human review - discrepancies
    # "HATFLDFL": HATFLDFL(),  # TODO: HATFLD family needs human review - discrepancies
    # "HATFLDFLS": HATFLDFLS(),  # TODO: HATFLD family needs human review
    # "HATFLDGLS": HATFLDGLS(),  # TODO: HATFLD family needs human review
    # "HEART6LS": HEART6LS(),  # TODO: Human review - significant discrepancies
    # "HEART8LS": HEART8LS(),  # TODO: Human review - significant discrepancies
    "HELIX": HELIX(),
    # "HIELOW": HIELOW(),  # TODO: Human review - significant discrepancies
    "HILBERTA": HILBERTA(),
    "HILBERTB": HILBERTB(),
    # "HIMMELBB": HIMMELBB(),  # TODO: Human review - Hessian discrepancies
    "HIMMELBCLS": HIMMELBCLS(),
    # "HIMMELBF": HIMMELBF(),  # TODO: Human review - Hessian discrepancies
    "HIMMELBG": HIMMELBG(),
    "HIMMELBH": HIMMELBH(),
    "HUMPS": HUMPS(),
    "INDEF": INDEF(),
    "INTEQNELS": INTEQNELS(),
    "JENSMP": JENSMP(),
    "JUDGE": JUDGE(),
    "KIRBY2LS": KIRBY2LS(),
    "KOWOSB": KOWOSB(),
    # "KSSLS": KSSLS(),  # TODO: Human review - significant obj/grad discrepancies
    "LANCZOS1LS": LANCZOS1LS(),
    "LANCZOS2LS": LANCZOS2LS(),
    "LIARWHD": LIARWHD(),
    "LSC1LS": LSC1LS(),
    "LSC2LS": LSC2LS(),
    # "MANCINO": MANCINO(),  # TODO: Human review - significant discrepancies in all
    "ROSENBR": ROSENBR(),
    "TENFOLDTRLS": TENFOLDTRLS(),
}


def get_problem(name: str):
    return problems_dict.get(name, None)  # TODO: try except with nicer error message


problems = (
    unconstrained_minimisation_problems
    + bounded_minimisation_problems
    + constrained_minimisation_problems
)



import jax
import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: human review required
class HATFLDD(AbstractUnconstrainedMinimisation):
    """An exponential fitting test problem from the OPTIMA user manual.

    Source:
    "The OPTIMA user manual (issue No.8, p. 35)",
    Numerical Optimization Centre, Hatfield Polytechnic (UK), 1989.

    SIF input: Ph. Toint, May 1990.

    Classification: SUR2-AN-3-0
    """

    def objective(self, y, args):
        del args
        x1, x2, x3 = y

        # Time points
        t_values = jnp.array([0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.75, 0.8, 0.85, 0.9])

        # Data values
        z_values = jnp.array(
            [1.751, 1.561, 1.391, 1.239, 1.103, 0.981, 0.925, 0.8721, 0.8221, 0.7748]
        )

        # Compute model values for each data point
        # The model is x1 * exp(t * x2) - exp(t * x3)
        model_values = x1 * jnp.exp(t_values * x2) - jnp.exp(t_values * x3)

        # Compute residuals
        residuals = model_values - z_values

        # Objective function is sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Initial point from SIF file (lines 75-77)
        return jnp.array([1.0, -1.0, 0.0])

    def args(self):
        return None

    def expected_result(self):
        # Not provided in the SIF file
        return None

    def expected_objective_value(self):
        # According to SIF file (line 120), the optimal objective value is 6.615114D-08
        return jnp.array(6.615114e-08)


# TODO: human review required
class HATFLDE(AbstractUnconstrainedMinimisation):
    """An exponential fitting test problem from the OPTIMA user manual.

    Source:
    "The OPTIMA user manual (issue No.8, p. 37)",
    Numerical Optimization Centre, Hatfield Polytechnic (UK), 1989.

    SIF input: Ph. Toint, May 1990.

    Classification: SUR2-AN-3-0
    """

    def objective(self, y, args):
        del args
        x1, x2, x3 = y

        # Time points
        t_values = jnp.array(
            [
                0.3,
                0.35,
                0.4,
                0.45,
                0.5,
                0.55,
                0.6,
                0.65,
                0.7,
                0.75,
                0.8,
                0.85,
                0.9,
                0.95,
                1.0,
                1.05,
                1.1,
                1.15,
                1.2,
                1.25,
                1.3,
            ]
        )

        # Data values
        z_values = jnp.array(
            [
                1.561,
                1.473,
                1.391,
                1.313,
                1.239,
                1.169,
                1.103,
                1.04,
                0.981,
                0.925,
                0.8721,
                0.8221,
                0.7748,
                0.73,
                0.6877,
                0.6477,
                0.6099,
                0.5741,
                0.5403,
                0.5084,
                0.4782,
            ]
        )

        # Compute model values for each data point
        # The model is x1 * exp(t * x2) - exp(t * x3)
        model_values = x1 * jnp.exp(t_values * x2) - jnp.exp(t_values * x3)

        # Compute residuals
        residuals = model_values - z_values

        # Objective function is sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Initial point from SIF file (lines 97-99)
        return jnp.array([1.0, -1.0, 0.0])

    def args(self):
        return None

    def expected_result(self):
        # Not provided in the SIF file
        return None

    def expected_objective_value(self):
        # According to SIF file (line 142), the optimal objective value is 5.120377D-07
        return jnp.array(5.120377e-07)


# TODO: human review required
class HATFLDFL(AbstractUnconstrainedMinimisation):
    """Fletcher's variation of a test problem (HATFLDF) from the OPTIMA user manual.

    Monotonic paths to the solution from the initial point move to infinity and back.

    Source:
    "The OPTIMA user manual (issue No.8, p. 47)",
    Numerical Optimization Centre, Hatfield Polytechnic (UK), 1989.

    SIF input: Ph. Toint, May 1990, mods Nick Gould, August 2008

    Nonlinear least-squares variant

    Classification: SUR2-AN-3-0
    """

    def objective(self, y, args):
        del args
        x1, x2, x3 = y

        # Constants from SIF file (lines 43-45)
        targets = jnp.array([0.032, 0.056, 0.099])

        # Parameters for the element uses
        t_values = jnp.array([1.0, 2.0, 3.0])

        # Compute model values for each data point
        # The model uses x1 + x2 * x3^t for each t in t_values
        # Note: XPEXP in this case is defined differently from HATFLDD and HATFLDE
        # Here, it's specifically x * y^t where t is an integer, not an exponential
        model_values = x1 + x2 * jnp.power(x3, t_values)

        # Compute residuals
        residuals = model_values - targets

        # Objective function is sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Initial point from SIF file (the "nasty" starting point, lines 55-57)
        return jnp.array([1.2, -1.2, 0.98])

    def args(self):
        return None

    def expected_result(self):
        # Not provided in the SIF file
        return None

    def expected_objective_value(self):
        # Not provided in the SIF file
        return None


# TODO: human review required
class HATFLDFLS(AbstractUnconstrainedMinimisation):
    """A test problem from the OPTIMA user manual.

    Least-squares version of HATFLDF.

    Source:
    "The OPTIMA user manual (issue No.8, p. 47)",
    Numerical Optimization Centre, Hatfield Polytechnic (UK), 1989.

    SIF input: Ph. Toint, May 1990.
    Least-squares version of HATFLDF.SIF, Nick Gould, Jan 2020.

    Classification: SUR2-AN-3-0
    """

    def objective(self, y, args):
        del args
        x1, x2, x3 = y

        # Constants from SIF file (lines 40-42)
        targets = jnp.array([0.032, 0.056, 0.099])

        # Parameters for the element uses
        t_values = jnp.array([1.0, 2.0, 3.0])

        # Compute model values for each data point
        # The model is x1 + x2 * exp(t * x3)
        model_values = x1 + x2 * jnp.exp(t_values * x3)

        # Compute residuals
        residuals = model_values - targets

        # Objective function is sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Initial point from SIF file (line 50)
        return jnp.full(3, 0.1)

    def args(self):
        return None

    def expected_result(self):
        # Not provided in the SIF file
        return None

    def expected_objective_value(self):
        # Not provided in the SIF file
        return None


# TODO: human review required
class HATFLDGLS(AbstractUnconstrainedMinimisation):
    """A test problem from the OPTIMA user manual.

    Least-squares version of HATFLDG.

    Source:
    "The OPTIMA user manual (issue No.8, p. 49)",
    Numerical Optimization Centre, Hatfield Polytechnic (UK), 1989.

    SIF input: Ph. Toint, May 1990.
    Least-squares version of HATFLDG.SIF, Nick Gould, Jan 2020.

    Classification: SUR2-AY-25-0
    """

    def objective(self, y, args):
        del args

        # Create residuals for each group
        # Each group is defined by G(i) = x(i) + x(13) - 1.0 for i=1...n
        residuals = y + y[12] - 1.0

        # Group element G(1) also involves A(1) with a -1.0 coefficient
        # A(1) is a 2PR element with x1 and x2, computing x1 * x2
        residuals = residuals.at[0].add(-1.0 * y[0] * y[1])

        # Groups G(2) to G(n) involve A(i) elements
        # For i=2 to n-1, A(i) is a 2PRI element with x(i), x(i-1), and x(i+1)
        # 2PRI computes (x(i) + 1.0) * (x(i-1) + 1.0 - x(i+1))

        # Use vmap instead of for-loop
        def compute_2pri_element(i):
            return (y[i] + 1.0) * (y[i - 1] + 1.0 - y[i + 1])

        indices = jnp.arange(1, len(y) - 1)
        inner_residuals = jax.vmap(compute_2pri_element)(indices)
        residuals = residuals.at[1:-1].add(inner_residuals)

        # For i=n, A(n) is a 2PR element with x(n-1) and x(n), computing x(n-1) * x(n)
        residuals = residuals.at[-1].add(y[-2] * y[-1])

        # Objective function is sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Initial point from SIF file (line 57)
        return jnp.ones(25)  # Hard-coded as 25 per SIF file

    def args(self):
        return None

    def expected_result(self):
        # Not provided in the SIF file
        return None

    def expected_objective_value(self):
        # Not provided in the SIF file
        return None



import jax
import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: Human review needed - gradient test fails
# The implementation computes the correct objective value but the gradient
# has discrepancies in the last few elements. This is likely due to subtle
# issues in how the banded structure interacts with the nonlinear elements
# in the lower right corner of the system.
class BRYBND(AbstractUnconstrainedMinimisation):
    """Broyden banded system of nonlinear equations.

    Source: problem 31 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley#73 (p. 41) and Toint#18
    SIF input: Ph. Toint, Dec 1989.

    Classification: SUR2-AN-V-0
    """

    n: int = 5000  # Dimension of the problem
    kappa1: float = 2.0  # Parameter
    kappa2: float = 5.0  # Parameter
    kappa3: float = 1.0  # Parameter
    lb: int = 5  # Lower bandwidth
    ub: int = 1  # Upper bandwidth

    def objective(self, y, args):
        del args
        n = self.n
        k1 = self.kappa1
        k2 = self.kappa2
        k3 = self.kappa3
        lb = self.lb
        ub = self.ub

        # Pre-compute powers of y for efficiency
        y_squared = y**2
        y_cubed = y**3

        # Create index array
        indices = jnp.arange(n)

        # Function to compute residual for a single index
        def compute_residual_i(i):
            # Linear part: k1 * x[i]
            res = k1 * y[i]

            # Determine region
            is_upper_left = i < lb
            is_lower_right = i >= n - ub

            # Compute band indices
            j_lower_start = jnp.maximum(0, i - lb)
            j_upper_end = jnp.minimum(n, i + ub + 1)

            # Create masks for bands
            j_indices = jnp.arange(n)
            lower_band_mask = (j_indices >= j_lower_start) & (j_indices < i)
            upper_band_mask = (j_indices > i) & (j_indices < j_upper_end)

            # Sum linear contributions from bands
            res = res - k3 * jnp.sum(y * lower_band_mask)
            res = res - k3 * jnp.sum(y * upper_band_mask)

            # The nonlinear terms are ADDED, not subtracted
            # And they come from elements E(j) and Q(j) in surrounding positions
            # not just at position i

            # For upper left corner (i < lb):
            #   - Add k2 * Q(i) at position i (cubed)
            #   - Add -k3 * E(j) for j < i and j > i (squared)
            if_upper_left = (
                k2 * y_cubed[i]
                - k3 * jnp.sum(y_squared * lower_band_mask)
                - k3 * jnp.sum(y_squared * upper_band_mask)
            )

            # For middle (lb <= i < n-ub):
            #   - Add k2 * E(i) at position i (squared)
            #   - Add -k3 * Q(j) for j < i (cubed)
            #   - Add -k3 * E(j) for j > i (squared)
            if_middle = (
                k2 * y_squared[i]
                - k3 * jnp.sum(y_cubed * lower_band_mask)
                - k3 * jnp.sum(y_squared * upper_band_mask)
            )

            # For lower right corner (i >= n-ub):
            #   - Add k2 * Q(i) at position i (cubed)
            #   - Add -k3 * E(j) for j < i and j > i (squared)
            if_lower_right = (
                k2 * y_cubed[i]
                - k3 * jnp.sum(y_squared * lower_band_mask)
                - k3 * jnp.sum(y_squared * upper_band_mask)
            )

            # Select appropriate nonlinear contributions
            nonlinear_term = jnp.where(
                is_upper_left,
                if_upper_left,
                jnp.where(is_lower_right, if_lower_right, if_middle),
            )
            res = res + nonlinear_term

            return res

        # Vectorize over all indices
        residuals = jax.vmap(compute_residual_i)(indices)

        # Return the sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Initial values from SIF file (all 1.0)
        return jnp.ones(self.n)

    def args(self):
        return None

    def expected_result(self):
        # The optimal solution is not explicitly provided in the SIF file
        return None

    def expected_objective_value(self):
        # According to the SIF file comment (line 213),
        # the optimal objective value is 0.0
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: Human review needed to verify the implementation matches the problem definition
class LIARWHD(AbstractUnconstrainedMinimisation):
    """Simplified version of the NONDIA problem.

    Source:
    G. Li,
    "The secant/finite difference algorithm for solving sparse
    nonlinear systems of equations",
    SIAM Journal on Optimization, (to appear), 1990.

    SIF input: Ph. Toint, Aug 1990.
    Classification: SUR2-AN-V-0
    """

    # Problem dimension
    n: int = 5000  # Other suggested values: 36, 100, 500, 1000, 10000

    def objective(self, y, args):
        """Compute the objective function value.

        The objective consists of two parts from AMPL:
        1. sum {i in 1..N} 4*(-x[1]+x[i]^2)^2
        2. sum {i in 1..N} (x[i]-1.0)^2
        """
        x1 = y[0]

        # First term: 4*(-x[1]+x[i]^2)^2 for all i
        first_terms = 4.0 * (-x1 + y**2) ** 2

        # Second term: (x[i]-1.0)^2 for all i
        second_terms = (y - 1.0) ** 2

        return jnp.sum(first_terms) + jnp.sum(second_terms)

    def y0(self):
        """Initial point with all variables set to 4.0."""
        return jnp.ones(self.n) * 4.0

    def args(self):
        """No additional arguments needed."""
        return None

    def expected_result(self):
        """The solution is all ones according to the problem definition."""
        return jnp.ones(self.n)

    def expected_objective_value(self):
        """The solution value is 0.0."""
        return jnp.array(0.0)



import jax
import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: needs human review
class GROWTHLS(AbstractUnconstrainedMinimisation):
    """GROWTHLS - Growth problem in 3 variables.

    This problem involves fitting the observed growth g(n) from Gaussian
    Elimination with complete pivoting to a function of the form:

    U1 * n ** (U2 + LOG(n) * U3)

    Source: NIST nonlinear regression test set

    SIF input: Nick Gould, Nov, 1991, modified by Ph. Toint, March 1994.

    Classification: SUR2-AN-3-0
    """

    def objective(self, y, args):
        del args

        # Extract the 3 parameters
        u1, u2, u3 = y

        # Data points: n values (input)
        n_values = jnp.array(
            [8.0, 9.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 18.0, 20.0, 25.0]
        )

        # Observed growth values g(n) (output)
        g_values = jnp.array(
            [
                8.4305,
                9.5294,
                10.4627,
                12.0,
                13.0205,
                14.5949,
                16.1078,
                18.0596,
                20.4569,
                24.25,
                32.9863,
            ]
        )

        # Add the missing value for n=8 (based on the constant section)
        g_values = jnp.array(
            [
                8.0,
                8.4305,
                9.5294,
                10.4627,
                12.0,
                13.0205,
                14.5949,
                16.1078,
                18.0596,
                20.4569,
                24.25,
                32.9863,
            ]
        )

        # Define the model function: u1 * n ** (u2 + log(n) * u3)
        def model_func(n):
            log_n = jnp.log(n)
            power = u2 + log_n * u3
            return u1 * n**power

        # Compute model predictions for all n values
        predictions = jax.vmap(model_func)(n_values)

        # Compute residuals
        residuals = predictions - g_values

        # Return sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Starting point from the SIF file: U1 = 100.0, U2 and U3 not specified
        # Using reasonable values for U2 and U3
        return jnp.array([100.0, 0.0, 0.0])

    def args(self):
        return None

    def expected_result(self):
        # The optimal solution is not specified in the SIF file
        return None

    def expected_objective_value(self):
        # The minimum objective value is not specified in the SIF file
        return None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HS24(AbstractConstrainedMinimisation):
    """Problem 24 from the Hock-Schittkowski test collection.

    A 2-variable nonlinear function with three inequality constraints and bounds.

    f(x) = (1/(27√3)) * ((x₁ - 3)² - 9) * x₂³

    Subject to: x₁/√3 - x₂ ≥ 0
                x₁ + √3*x₂ ≥ 0
                -x₁ - √3*x₂ + 6 ≥ 0
                0 ≤ x₁
                0 ≤ x₂

    Source: problem 24 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    Also cited: Betts [8], Box [12]

    Classification: PLR-T1-1
    """

    def objective(self, y, args):
        x1, x2 = y
        sqrt3 = jnp.sqrt(3.0)
        return (1.0 / (27.0 * sqrt3)) * ((x1 - 3.0) ** 2 - 9.0) * (x2**3)

    def y0(self):
        return jnp.array([1.0, 0.5])

    def args(self):
        return None

    def expected_result(self):
        sqrt3 = jnp.sqrt(3.0)
        return jnp.array([3.0, sqrt3])

    def expected_objective_value(self):
        return jnp.array(-1.0)

    def bounds(self):
        return (jnp.array([0.0, 0.0]), jnp.array([jnp.inf, jnp.inf]))

    def constraint(self, y):
        x1, x2 = y
        sqrt3 = jnp.sqrt(3.0)
        # Inequality constraints
        inequality_constraint = jnp.array(
            [
                x1 / sqrt3 - x2,  # x₁/√3 - x₂ ≥ 0
                x1 + sqrt3 * x2,  # x₁ + √3*x₂ ≥ 0
                -x1 - sqrt3 * x2 + 6.0,  # -x₁ - √3*x₂ + 6 ≥ 0
            ]
        )
        return None, inequality_constraint



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HS23(AbstractConstrainedMinimisation):
    """Problem 23 from the Hock-Schittkowski test collection.

    A 2-variable quadratic function with five inequality constraints and bounds.

    f(x) = x₁² + x₂²

    Subject to: x₁ + x₂ - 1 ≥ 0
                x₁² + x₂² - 1 ≥ 0
                9x₁² + x₂² - 9 ≥ 0
                x₁² - x₂ ≥ 0
                x₂² - x₁ ≥ 0
                -50 ≤ x₁ ≤ 50
                -50 ≤ x₂ ≤ 50

    Source: problem 23 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    Also cited: Betts [8]

    Classification: QQR-T1-7
    """

    def objective(self, y, args):
        x1, x2 = y
        return x1**2 + x2**2

    def y0(self):
        return jnp.array([3.0, 1.0])

    def args(self):
        return None

    def expected_result(self):
        return jnp.array([1.0, 1.0])

    def expected_objective_value(self):
        return jnp.array(2.0)

    def bounds(self):
        return (jnp.array([-50.0, -50.0]), jnp.array([50.0, 50.0]))

    def constraint(self, y):
        x1, x2 = y
        # Inequality constraints
        inequality_constraint = jnp.array(
            [
                x1 + x2 - 1,  # x₁ + x₂ - 1 ≥ 0
                x1**2 + x2**2 - 1,  # x₁² + x₂² - 1 ≥ 0
                9 * x1**2 + x2**2 - 9,  # 9x₁² + x₂² - 9 ≥ 0
                x1**2 - x2,  # x₁² - x₂ ≥ 0
                x2**2 - x1,  # x₂² - x₁ ≥ 0
            ]
        )
        return None, inequality_constraint



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class BT7(AbstractConstrainedMinimisation):
    """BT7 - Boggs-Tolle test problem 7.

    n = 5, m = 3.
    f(x) = 100(x₂ - x₁²)² + (1 - x₁)².
    g₁(x) = x₁x₂ - 1 - x₅.
    g₂(x) = x₂² + x₁ - x₃².
    g₃(x) = x₁ + x₅² - ½.

    Start 1: x₁ = -2, x₂ = 1, x₃ = 1, x₄ = 1, x₅ = 1.
    Start 2: x₁ = -20, x₂ = 10, x₃ = 1, x₄ = 1, x₅ = 1.
    Solution: x* = (-0.79212, -1.2624, 0.0, -0.89532, 1.1367).

    Source: Boggs, P.T. and Tolle, J.W.,
    "A strategy for global convergence in a sequential
    quadratic programming algorithm",
    SIAM J. Numer. Anal. 26(3), pp. 600-623, 1989.

    Classification: OQR2-AN-5-3
    """

    y0_id: int = 0
    provided_y0s: frozenset = frozenset({0, 1})

    def objective(self, y, args):
        del args
        x1, x2, x3, x4, x5 = y
        return 100 * (x2 - x1**2) ** 2 + (1 - x1) ** 2

    def y0(self):
        if self.y0_id == 0:
            return jnp.array([-2.0, 1.0, 1.0, 1.0, 1.0])
        elif self.y0_id == 1:
            return jnp.array([-20.0, 10.0, 1.0, 1.0, 1.0])

    def args(self):
        return None

    def expected_result(self):
        return jnp.array([-0.79212, -1.2624, 0.0, -0.89532, 1.1367])

    def expected_objective_value(self):
        return None  # Not explicitly given

    def bounds(self):
        return None

    def constraint(self, y):
        x1, x2, x3, x4, x5 = y
        # Equality constraints
        g1 = x1 * x2 - 1 - x5
        g2 = x2**2 + x1 - x3**2
        g3 = x1 + x5**2 - 0.5
        equality_constraints = jnp.array([g1, g2, g3])
        return equality_constraints, None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HS71(AbstractConstrainedMinimisation):
    """Problem 71 from the Hock-Schittkowski test collection.

    A 4-variable nonlinear objective function with one equality constraint, one
    inequality constraint and bounds.

    f(x) = x₁*x₄*(x₁ + x₂ + x₃) + x₃

    Subject to:
        x₁*x₂*x₃*x₄ - 25 ≥ 0
        x₁² + x₂² + x₃² + x₄² - 40 = 0
        1 ≤ xᵢ ≤ 5, i=1,...,4

    Source: problem 71 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    Also cited: Bartholomew-Biggs [4]

    Classification: PPR-P1-3
    """

    def objective(self, y, args):
        x1, x2, x3, x4 = y
        return x1 * x4 * (x1 + x2 + x3) + x3

    def y0(self):
        return jnp.array([1.0, 5.0, 5.0, 1.0])  # feasible according to the problem

    def args(self):
        return None

    def expected_result(self):
        return jnp.array([1.0, 4.742999, 3.821150, 1.379408])

    def expected_objective_value(self):
        return jnp.array(17.0140173)

    def bounds(self):
        return (jnp.array([1.0, 1.0, 1.0, 1.0]), jnp.array([5.0, 5.0, 5.0, 5.0]))

    def constraint(self, y):
        x1, x2, x3, x4 = y
        # Inequality constraint (g(x) ≥ 0)
        ineq1 = x1 * x2 * x3 * x4 - 25
        # Equality constraint
        eq1 = x1**2 + x2**2 + x3**2 + x4**2 - 40

        equality_constraints = jnp.array([eq1])
        inequality_constraints = jnp.array([ineq1])
        return equality_constraints, inequality_constraints



import jax
import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class INTEQNELS(AbstractUnconstrainedMinimisation):
    """The discrete integral equation function in least-squares form.

    This is problem 29 from:
    J.J. Moré, B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    The problem discretizes an integral equation. The function f_i(x) is defined as:
    f_i(x) = x_i + h[(1-t_i)∑_{j=0}^i t_j(x_j + t_j + 1)³
        + t_i ∑_{j=i+1}^{n-1} (1-t_j)(x_j + t_j + 1)³]/2

    where h = 1/(n-1), t_i = ih for i = 0, ..., n-1.
    The objective is the sum of squares: ∑_{i=0}^{n-1} f_i(x)².

    Initial point: x_i = t_i(t_i - 1) for i = 0, ..., n-1.

    SIF input: Ph. Toint, Feb 1990.
    Modification to remove fixed variables: Nick Gould, Oct 2015.
    Classification: SUR2-AN-V-0
    """

    n: int = 502  # Default value to match pycutest

    def objective(self, y, args):
        """Compute the objective function value as sum of squares of f_i(x)."""
        n = self.n
        h = 1.0 / (n - 1)  # Match the starting point indexing

        # t_i = ih for i = 0, 1, ..., n-1 (to match starting point)
        t = jnp.arange(n) * h

        # y represents x_0, x_1, ..., x_{n-1}
        x = y

        def compute_f_i(i):
            """Compute f_i(x) for the i-th equation (0-indexed)."""
            t_i = t[i]
            x_i = x[i]

            # Use masking for both sums to avoid dynamic arange
            all_j = jnp.arange(n)
            t_j = t[all_j]
            x_j = x[all_j]

            # First sum: ∑_{j=0}^i t_j(x_j + t_j + 1)³
            first_mask = all_j <= i
            first_values = jnp.where(first_mask, t_j * (x_j + t_j + 1.0) ** 3, 0.0)
            first_sum = jnp.sum(jnp.asarray(first_values))

            # Second sum: ∑_{j=i+1}^{n-1} (1-t_j)(x_j + t_j + 1)³
            second_mask = all_j > i
            second_values = jnp.where(
                second_mask, (1.0 - t_j) * (x_j + t_j + 1.0) ** 3, 0.0
            )
            second_sum = jnp.sum(jnp.asarray(second_values))

            # Compute f_i(x)
            integral_part = h * ((1.0 - t_i) * first_sum + t_i * second_sum) / 2.0
            f_i = x_i + integral_part

            return f_i

        # Compute all f_i values and sum their squares
        i_values = jnp.arange(n)
        f_values = jax.vmap(compute_f_i)(i_values)

        return jnp.sum(f_values**2)

    def y0(self):
        """Initial point: ξ_i = t_i(t_i - 1) where t_i = ih.

        From the reference: x_0 = (ξ_i) where ξ_i = t_j(t_j - 1).

        Based on PyCUTEst output, it seems to use i = 0, 1, ..., n-1 with h = 1/(n-1).
        This gives t_i = ih and ξ_i = t_i(t_i - 1) starting with ξ_0 = 0.
        """
        n = self.n
        h = 1.0 / (n - 1)  # This matches PyCUTEst behavior
        # For i = 0, 1, 2, ..., n-1, compute t_i = i*h and ξ_i = t_i*(t_i - 1)
        i_vals = jnp.arange(n)
        t_vals = i_vals * h
        xi_vals = t_vals * (t_vals - 1.0)
        return xi_vals

    def args(self):
        """No additional arguments needed."""
        return None

    def expected_result(self):
        """The solution is not specified in the SIF file."""
        return None

    def expected_objective_value(self):
        """The solution value is not specified in the SIF file."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HS36(AbstractConstrainedMinimisation):
    """Problem 36 from the Hock-Schittkowski test collection.

    A 3-variable nonlinear function with one inequality constraint and bounds.

    f(x) = -x₁x₂x₃

    Subject to: 72 - x₁ - 2x₂ - 2x₃ ≥ 0
                0 ≤ x₁ ≤ 20
                0 ≤ x₂ ≤ 11
                0 ≤ x₃ ≤ 42

    Source: problem 36 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    Also cited: Biggs [10]

    Classification: PLR-T1-2
    """

    def objective(self, y, args):
        x1, x2, x3 = y
        return -x1 * x2 * x3

    def y0(self):
        return jnp.array([10.0, 10.0, 10.0])

    def args(self):
        return None

    def expected_result(self):
        return jnp.array([20.0, 11.0, 15.0])

    def expected_objective_value(self):
        return jnp.array(-3300.0)

    def bounds(self):
        return (jnp.array([0.0, 0.0, 0.0]), jnp.array([20.0, 11.0, 42.0]))

    def constraint(self, y):
        x1, x2, x3 = y
        # Inequality constraint: 72 - x₁ - 2x₂ - 2x₃ ≥ 0
        inequality_constraint = jnp.array([72.0 - x1 - 2.0 * x2 - 2.0 * x3])
        return None, inequality_constraint



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: This implementation requires human review and verification against
# another CUTEst interface
class FLETCHBV(AbstractUnconstrainedMinimisation):
    """The FLETCHBV function.

    Another Boundary Value problem.

    Source: The first problem given by
    R. Fletcher, "An optimal positive definite update for sparse Hessian matrices"
    Numerical Analysis report NA/145, University of Dundee, 1992.

    N.B. This formulation is incorrect. See FLETCBV2.SIF for
    the correct version.

    SIF input: Nick Gould, Oct 1992.
    Classification: OUR2-AN-V-0
    """

    n: int = 5000  # Default dimension in SIF file
    kappa: float = 1.0  # Parameter used in the problem
    objscale: float = 1.0  # Object scaling parameter

    def objective(self, y, args):
        del args
        h = 1.0 / (self.n + 1)
        h2 = h * h
        p = 1.0 / self.objscale  # objscale = 1.0 for FLETCHBV

        # From AMPL file:
        # 0.5*p*(x[1])^2 +
        # sum {i in 1..n-1} 0.5*p*(x[i]-x[i+1])^2 +
        # 0.5*p*(x[n])^2 +
        # sum {i in 1..n} (p*(-1-2/h^2)*x[i]) +
        # sum {i in 1..n} (-kappa*p*cos(x[i])/h^2);

        term1 = 0.5 * p * (y[0]) ** 2
        term2 = 0.5 * p * jnp.sum((y[:-1] - y[1:]) ** 2)
        term3 = 0.5 * p * (y[-1]) ** 2
        term4 = p * (-1.0 - 2.0 / h2) * jnp.sum(y)
        term5 = -self.kappa * p / h2 * jnp.sum(jnp.cos(y))

        return term1 + term2 + term3 + term4 + term5

    def y0(self):
        # Initial values from SIF file: i*h for i=1..n
        h = 1.0 / (self.n + 1)
        return jnp.arange(1, self.n + 1) * h

    def args(self):
        return None

    def expected_result(self):
        return None

    def expected_objective_value(self):
        return None


# TODO: This implementation requires human review and verification against
# another CUTEst interface
class FLETBV3M(AbstractUnconstrainedMinimisation):
    """The FLETBV3M function.

    Variant of FLETCBV3, another boundary value problem, by Luksan et al

    Source: problem 30 in
    L. Luksan, C. Matonoha and J. Vlcek
    Modified CUTE problems for sparse unconstrained optimization
    Technical Report 1081
    Institute of Computer Science
    Academy of Science of the Czech Republic

    based on a scaled version of the first problem given by
    R. Fletcher,
    "An optimal positive definite update for sparse Hessian matrices"
    Numerical Analysis report NA/145, University of Dundee, 1992.

    SIF input: Nick Gould, June, 2013
    Classification: OUR2-AN-V-0
    """

    n: int = 5000  # Default dimension in SIF file
    kappa: float = 1.0  # Parameter used in the problem
    objscale: float = 1.0e8  # Scaling factor

    def objective(self, y, args):
        del args
        h = 1.0 / (self.n + 1)
        h2 = h * h
        p = 1.0 / self.objscale

        # Define each term based on the SIF file
        # G(0): p * 0.5 * (x_1)^2
        f1 = 0.5 * p * (y[0]) ** 2

        # G(i) for i=1...n-1: p * 0.5 * (x_i - x_{i+1})^2
        f2 = 0.5 * p * jnp.sum((y[:-1] - y[1:]) ** 2)

        # G(n): p * 0.5 * (x_n)^2
        f3 = 0.5 * p * (y[-1]) ** 2

        # C(i): p * cos(x_i) * (-kappa/h^2)
        f4 = p * (-self.kappa / h2) * jnp.sum(jnp.cos(y))

        # S(i): 100 * sin(0.01 * x_i) * p * (1+2/h^2) - fix sign like FLETCBV3
        f5 = p * (1.0 + 2.0 / h2) * jnp.sum(100.0 * jnp.sin(0.01 * y))

        return f1 + f2 + f3 + f4 + f5

    def y0(self):
        # Initial values from SIF file: i*h for i=1..n
        h = 1.0 / (self.n + 1)
        return jnp.arange(1, self.n + 1) * h

    def args(self):
        return None

    def expected_result(self):
        return None

    def expected_objective_value(self):
        return None


# TODO: This implementation requires human review and verification against
# another CUTEst interface
class FLETCBV2(AbstractUnconstrainedMinimisation):
    """The FLETCBV2 function.

    Another Boundary Value problem.

    Source: The first problem given by
    R. Fletcher, "An optimal positive definite update for sparse Hessian matrices"
    Numerical Analysis report NA/145, University of Dundee, 1992.
    but assuming that the 1/h**2 term should read h**2
    This is what Fletcher intended (private communication).

    The author comments: "The problem arises from discretizing the bvp
                 x"=-2+sin x in [0,1]
    with x(0)=0, x(1)=1. This gives a symmetric system of equations,
    the residual vector of which is the gradient of the given function."
    He multiplies through by h^2 before integrating.

    SIF input: Nick Gould, Nov 1992.
    Classification: OUR2-AN-V-0
    """

    n: int = 5000  # Default dimension in SIF file
    kappa: float = 1.0  # Parameter used in the problem

    def objective(self, y, args):
        del args
        h = 1.0 / (self.n + 1)
        h2 = h * h

        # Define components as described in the SIF file
        # G(0): 0.5 * (x_1)^2
        f1 = 0.5 * (y[0]) ** 2

        # G(i) for i=1...n-1: 0.5 * (x_i - x_{i+1})^2
        f2 = 0.5 * jnp.sum((y[:-1] - y[1:]) ** 2)

        # G(n): 0.5 * (x_n)^2
        f3 = 0.5 * (y[-1]) ** 2

        # L(i) for i=1...n-1: x_i * (-2*h2)
        f4 = -2.0 * h2 * jnp.sum(y[:-1])

        # L(n): x_n * (-1-2*h2)
        f5 = (-1.0 - 2.0 * h2) * y[-1]

        # C(i): -kappa*h2 * cos(x_i)
        f6 = -self.kappa * h2 * jnp.sum(jnp.cos(y))

        return f1 + f2 + f3 + f4 + f5 + f6

    def y0(self):
        # Initial values from SIF file: i*h for i=1..n
        h = 1.0 / (self.n + 1)
        return jnp.arange(1, self.n + 1) * h

    def args(self):
        return None

    def expected_result(self):
        return None

    def expected_objective_value(self):
        return None


# TODO: This implementation requires human review and verification against
# another CUTEst interface
class FLETCHCR(AbstractUnconstrainedMinimisation):
    """The FLETCHCR function.

    The chained Rosenbrock function as given by Fletcher.

    Source: The second problem given by
    R. Fletcher, "An optimal positive definite update for sparse Hessian matrices"
    Numerical Analysis report NA/145, University of Dundee, 1992.

    SIF input: Nick Gould, Oct 1992.
    Classification: OUR2-AN-V-0
    """

    n: int = 1000  # Default dimension in SIF file

    def objective(self, y, args):
        del args

        # Let me try the exact AMPL formulation without the +1 term first
        # Standard chained Rosenbrock: sum 100*(x[i+1] - x[i]^2)^2 + (x[i] - 1)^2
        term1 = jnp.sum(100 * (y[1:] - y[:-1] ** 2) ** 2)
        term2 = jnp.sum((y[:-1] - 1.0) ** 2)

        return term1 + term2

    def y0(self):
        # Starting point: all zeros
        return jnp.zeros(self.n)

    def args(self):
        return None

    def expected_result(self):
        # Minimum value is 0, achieved when all variables are 1
        return jnp.ones(self.n)

    def expected_objective_value(self):
        return jnp.array(0.0)


# TODO: this has not yet been compared against another interface to CUTEst
class FLETCBV3(AbstractUnconstrainedMinimisation):
    """The FLETCBV3 function.

    A boundary value problem from Fletcher (1992).

    Source: The first problem given by
    R. Fletcher, "An optimal positive definite update for sparse Hessian matrices"
    Numerical Analysis report NA/145, University of Dundee, 1992.

    Note J. Haffner --------------------------------------------------------------------
    The reference given appears to be incorrect, the PDF available under the title above
    does not include a problem description.

    This can be defined for different dimensions (original SIF allows 10, 100, 1000,
    5000, or 10000), with 5000 being the default in the SIF file.
    ------------------------------------------------------------------------------------

    SIF input: Nick Gould, Oct 1992.
    Classification: OUR2-AN-V-0
    """

    n: int = 5000  # Default dimension in SIF file
    scale: float = 1e8  # Called OBJSCALE in the SIF file
    extra_term: int = 1  # Corresponds to the parameter kappa, which is 1 or 0

    def objective(self, y, args):
        p, kappa = args
        h = 1.0 / (self.n + 1)
        h2 = h * h

        # From AMPL: put p scaling back and try flipping the sign on linear term
        term1 = 0.5 * p * (y[0]) ** 2
        term2 = jnp.sum(0.5 * p * (y[:-1] - y[1:]) ** 2)
        term3 = 0.5 * p * (y[-1]) ** 2
        term4 = jnp.sum(p * (1.0 + 2.0 / h2) * y)  # Note: sign flipped
        term5 = jnp.sum(-kappa * p * jnp.cos(y) / h2)

        return term1 + term2 + term3 + term4 + term5

    def y0(self):
        n = self.n
        h = 1.0 / (self.n + 1)
        # Starting point according to SIF file: i*h for i=1..n
        return jnp.arange(1, n + 1) * h

    def args(self):
        # p and kappa from SIF file
        p = 1.0 / self.scale
        kappa = float(self.extra_term)
        return jnp.array([p, kappa])

    def expected_result(self):
        return None

    def expected_objective_value(self):
        return None  # Takes different values for different problem configurations



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: This implementation requires human review and verification against
# another CUTEst interface
class CUBE(AbstractUnconstrainedMinimisation):
    """A cubic variant of the Rosenbrock test function.

    This problem is a modification of the Rosenbrock function, using
    a cubic term instead of a quadratic. The objective function includes
    terms with (x_i - x_{i-1}^3)^2.

    Source: problem 5 (p. 89) in
    A.R. Buckley,
    "Test functions for unconstrained minimization",
    TR 1989CS-3, Mathematics, statistics and computing centre,
    Dalhousie University, Halifax (CDN), 1989.

    SIF input: Ph. Toint, Dec 1989.

    Classification: SUR2-AN-2-0
    """

    n: int = 2  # Number of variables

    def objective(self, y, args):
        del args

        # First term: (x_1 - 1)^2
        term1 = (y[0] - 1.0) ** 2

        # Second term: 100 * (x_2 - x_1^3)^2
        term2 = 100.0 * (y[1] - y[0] ** 3) ** 2

        return term1 + term2

    def y0(self):
        # Initial values from SIF file
        return jnp.array([-1.2, 1.0])

    def args(self):
        return None

    def expected_result(self):
        # Based on the optimal objective value of 0.0,
        # the solution must satisfy:
        # x_1 = 1
        # x_2 = x_1^3 = 1
        return jnp.array([1.0, 1.0])

    def expected_objective_value(self):
        # According to the SIF file, line 91
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class LUKVLE4C(AbstractConstrainedMinimisation):
    """LUKVLE4C - Corrected chained Cragg-Levy function with tridiagonal constraints.

    This is the corrected version of LUKVLE4, fixing the bug where group C(I) was
    defined twice in the original SIF file. The corrected version uses E(I) for the
    tan^4 terms in the objective function.

    TODO: Human review needed - 3% discrepancy with pycutest
    This implementation uses the corrected LUKVLE4C.SIF file which fixes a bug in
    the original LUKVLE4.SIF. While our implementation follows the corrected SIF
    file exactly, there remains a ~3% discrepancy with pycutest (3,020,250.61 vs
    3,112,445.79). Most gradient elements match exactly, but a few near the end
    differ. This may be due to numerical precision differences in tangent
    calculations or subtle SIFDECODE interpretation differences. The corrected
    version is vastly superior to the buggy LUKVLE4 (which has a factor of ~2380
    error), but does not yet achieve exact numerical agreement with pycutest.

    Problem 5.4 from Luksan and Vlcek test problems.

    The objective is a chained Cragg-Levy function:
    f(x) = Σ[i=1 to n/2-1] [(exp(x_{2i-1}) - x_{2i})^4 + 100(x_{2i} - x_{2i+1})^6 +
                            tan^4(x_{2i+1} - x_{2i+2}) + x_{2i-1}^8 + (x_{2i+2} - 1)^2]

    Subject to equality constraints:
    c_k(x) = 8x_{k+1}(x_{k+1}^2 - x_k) - 2(1 - x_{k+1}) + 4(x_{k+1} - x_{k+2}^2) = 0,
    for k = 1, ..., n-2

    Starting point:
    x_i = 1 for i ≡ 1 (mod 4)
    x_i = 2 for i ≢ 1 (mod 4)

    Source: L. Luksan and J. Vlcek,
    "Sparse and partially separable test problems for
    unconstrained and equality constrained optimization",
    Technical Report 767, Inst. Computer Science, Academy of Sciences
    of the Czech Republic, 182 07 Prague, Czech Republic, 1999

    SIF input: Nick Gould, April 2001
    Modified version of LUKVLE4 (formulation corrected) May 2024

    Classification: OOR2-AY-V-V
    """

    n: int = 10000  # Default dimension, can be overridden

    def objective(self, y, args):
        del args
        n = len(y)
        # Chained Cragg-Levy function - vectorized
        num_complete_groups = (n - 2) // 2  # Ensure we have all 4 elements
        if num_complete_groups <= 0 or n < 4:
            return jnp.array(0.0)

        i = jnp.arange(num_complete_groups)
        x_2i_minus_1 = y[2 * i]  # x_{2i-1} in 1-based
        x_2i = y[2 * i + 1]  # x_{2i} in 1-based
        x_2i_plus_1 = y[2 * i + 2]  # x_{2i+1} in 1-based
        x_2i_plus_2 = y[2 * i + 3]  # x_{2i+2} in 1-based

        # Compute all terms vectorized
        # Note: The SIF file may have issues at boundaries. Let's match exactly.
        terms = (
            (jnp.exp(x_2i_minus_1) - x_2i) ** 4
            + 100 * (x_2i - x_2i_plus_1) ** 6
            + jnp.tan(x_2i_plus_1 - x_2i_plus_2) ** 4
            + x_2i_minus_1**8
            + (x_2i_plus_2 - 1) ** 2
        )

        return jnp.sum(terms)

    def y0(self):
        # Starting point
        y = jnp.zeros(self.n)
        # x_i = 1 for i ≡ 1 (mod 4) -> 0-based: i ≡ 0 (mod 4)
        y = y.at[::4].set(1.0)
        # x_i = 2 for i ≢ 1 (mod 4) -> 0-based: i ≢ 0 (mod 4)
        y = y.at[1::4].set(2.0)
        y = y.at[2::4].set(2.0)
        y = y.at[3::4].set(2.0)
        return y

    def args(self):
        return None

    def expected_result(self):
        # Solution pattern based on problem structure
        return None  # Unknown exact solution

    def expected_objective_value(self):
        return None  # Unknown exact objective value

    def bounds(self):
        return None

    def constraint(self, y):
        n = len(y)
        if n < 3:
            return jnp.array([]), None

        # Constraints from k=1 to n-2
        # Vectorized constraint computation
        # For k = 0 to n-3, we need y[k], y[k+1], y[k+2]
        y_k = y[:-2]  # y[0] to y[n-3]
        y_k1 = y[1:-1]  # y[1] to y[n-2]
        y_k2 = y[2:]  # y[2] to y[n-1]

        # c_k: 8x_{k+1}(x_{k+1}^2 - x_k) - 2(1 - x_{k+1}) + 4(x_{k+1} - x_{k+2}^2)
        constraints = 8 * y_k1 * (y_k1**2 - y_k) - 2 * (1 - y_k1) + 4 * (y_k1 - y_k2**2)

        return constraints, None



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: human review required
class GULF(AbstractUnconstrainedMinimisation):
    """The Gulf RD test problem in 3 variables.

    This function is a nonlinear least squares with 99 groups. Each
    group has a nonlinear element of exponential type.

    Source: problem 11 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley#27
    SIF input: Ph. Toint, Dec 1989.

    Classification: SUR2-MN-3-0
    """

    n: int = 3  # Number of variables is 3
    m: int = 99  # Number of data points

    def objective(self, y, args):
        del args
        x1, x2, x3 = y

        # Create array of t_i values: t_i = 0.01 * i for i=1...m
        t_values = jnp.arange(1, self.m + 1) * 0.01

        # Compute the nonlinear expressions as defined in the SIF file (lines 133-142)
        # y_i terms: 25 + (-50 * log(t_i))^(2/3)
        y_i = 25.0 + jnp.power(-50.0 * jnp.log(t_values), 2.0 / 3.0)

        # y_i - x2 term
        y_minus_v2 = y_i - x2

        # Calculate |y_i - v_2|^v_3 / (-v_1) as in AMPL
        a_i = jnp.power(jnp.abs(y_minus_v2), x3) / (-x1)

        # Compute exp(a_i) - t_i as the residuals (corrected formula from AMPL)
        residuals = jnp.exp(a_i) - t_values

        # Sum of squares of residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Initial point from SIF file
        return jnp.array([5.0, 2.5, 0.15])

    def args(self):
        return None

    def expected_result(self):
        # Not provided in the SIF file
        return None

    def expected_objective_value(self):
        # According to SIF file (line 100), the optimal objective value is 0.0
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HS60(AbstractConstrainedMinimisation):
    """Problem 60 from the Hock-Schittkowski test collection.

    A 3-variable nonlinear objective function with one equality constraint and bounds.

    f(x) = (x₁ - 1)² + (x₁ - x₂)² + (x₂ - x₃)⁴

    Subject to:
        x₁(1 + x₂²) + x₃⁴ - 4 - 3√2 = 0
        -10 ≤ xᵢ ≤ 10, i=1,2,3

    Source: problem 60 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    Also cited: Betts [8], Miele e.al. [42,44]

    Classification: PPR-P1-1
    """

    def objective(self, y, args):
        x1, x2, x3 = y
        return (x1 - 1) ** 2 + (x1 - x2) ** 2 + (x2 - x3) ** 4

    def y0(self):
        return jnp.array([2.0, 2.0, 2.0])  # not feasible according to the problem

    def args(self):
        return None

    def expected_result(self):
        return jnp.array([1.10485902, 1.19667419, 1.53526226])

    def expected_objective_value(self):
        return jnp.array(0.03256820025)

    def bounds(self):
        return (jnp.array([-10.0, -10.0, -10.0]), jnp.array([10.0, 10.0, 10.0]))

    def constraint(self, y):
        x1, x2, x3 = y
        # Equality constraint
        eq1 = x1 * (1 + x2**2) + x3**4 - 4 - 3 * jnp.sqrt(2)
        equality_constraints = jnp.array([eq1])
        return equality_constraints, None



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: Needs verification against another CUTEst interface
class HIMMELBG(AbstractUnconstrainedMinimisation):
    """The HIMMELBG function.

    A 2-variable problem by Himmelblau.

    Source: problem 33 in
    D.H. Himmelblau,
    "Applied nonlinear programming",
    McGraw-Hill, New-York, 1972.
    See Buckley#87 (p. 67)

    SIF input: Ph. Toint, Dec 1989.
    Classification: OUR2-AN-2-0
    """

    def objective(self, y, args):
        del args
        x1, x2 = y
        exponential = jnp.exp(-x1 - x2)
        quadratic = 2 * x1**2 + 3 * x2**2
        return exponential * quadratic

    def y0(self):
        return jnp.array([0.5, 0.5])

    def args(self):
        return None

    def expected_result(self):
        return None

    def expected_objective_value(self):
        return jnp.array(0.0)


# TODO: human review required
class HIMMELBB(AbstractUnconstrainedMinimisation):
    """The HIMMELBB function.

    A 2-variable problem by Himmelblau.

    Source: problem 33 in
    D.H. Himmelblau,
    "Applied nonlinear programming",
    McGraw-Hill, New-York, 1972.

    SIF input: Ph. Toint, Dec 1989.
    Classification: OUR2-AN-2-0
    """

    def objective(self, y, args):
        del args
        x1, x2 = y
        term1 = x1 * x2 * (1 - x1)
        term2 = 1 - x2 - (x1 * (1 - x1) ** 5)
        return (term1 * term2) ** 2

    def y0(self):
        return jnp.array([-1.2, 1.0])

    def args(self):
        return None

    def expected_result(self):
        return None

    def expected_objective_value(self):
        return None


# TODO: human review required
class HIMMELBCLS(AbstractUnconstrainedMinimisation):
    """Himmelblau's nonlinear least-squares problem.

    A 2-variable problem with 2 residuals.

    Source: problem 201 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lecture Notes in Economics and Mathematical Systems 187,
    Springer Verlag, Berlin, 1981.

    SIF input: Ph. Toint, Dec 1989.
    Classification: SUR2-AN-2-0
    """

    def objective(self, y, args):
        del args
        x1, x2 = y

        # From SIF file:
        # G1 = X2 + X1² - 11.0
        # G2 = X1 + X2² - 7.0
        # Objective = G1² + G2²
        g1 = x2 + x1**2 - 11.0
        g2 = x1 + x2**2 - 7.0

        # Return the sum of squares
        return g1**2 + g2**2

    def y0(self):
        return jnp.array([1.0, 1.0])

    def args(self):
        return None

    def expected_result(self):
        # The exact solution is (±√7, ±√11)
        # We'll return one of the four solutions
        return jnp.array([jnp.sqrt(7.0), jnp.sqrt(11.0)])

    def expected_objective_value(self):
        # At the optimal solution, both residuals are 0
        return jnp.array(0.0)


class HIMMELBF(AbstractUnconstrainedMinimisation):
    """Himmelblau's HIMMELBF function.

    A 4-variable least squares data fitting problem.

    Source: problem 33 in
    D.H. Himmelblau,
    "Applied nonlinear programming",
    McGraw-Hill, New-York, 1972.

    SIF input: Ph. Toint, Dec 1989.
    Classification: SUR2-AN-4-0
    """

    def objective(self, y, args):
        del args
        x1, x2, x3, x4 = y

        # Data points provided in the SIF file
        a_values = jnp.array(
            [0.0, 0.000428, 0.001000, 0.001610, 0.002090, 0.003480, 0.005250]
        )
        b_values = jnp.array([7.391, 11.18, 16.44, 16.20, 22.20, 24.02, 31.32])

        # Compute the residuals for each data point
        def residual_fn(a, b):
            numerator = x1**2 + a * x2**2 + a**2 * x3**2
            denominator = b * (1 + a * x4**2)
            return (numerator / denominator) - 1.0

        # Vectorize the computation across all data points
        residuals = jnp.array([residual_fn(a, b) for a, b in zip(a_values, b_values)])

        # Return the sum of squared residuals
        # Scale the residuals: correct scaling is 1e4 (not 10e4 which is 10*1e4)
        return 1e4 * jnp.sum(residuals**2)

    def y0(self):
        return jnp.array([2.7, 90.0, 1500.0, 10.0])

    def args(self):
        return None

    def expected_result(self):
        return None

    def expected_objective_value(self):
        return None


# TODO: human review required
class HIMMELBH(AbstractUnconstrainedMinimisation):
    """Himmelblau's HIMMELBH function.

    A 2-variable problem.

    Source: problem 33 in
    D.H. Himmelblau,
    "Applied nonlinear programming",
    McGraw-Hill, New-York, 1972.

    SIF input: Ph. Toint, Dec 1989.
    Classification: OUR2-AN-2-0
    """

    def objective(self, y, args):
        del args
        x1, x2 = y

        term1 = x1**3 - 3 * x1
        term2 = x2**2 - 2 * x2
        constant = 2

        return term1 + term2 + constant

    def y0(self):
        return jnp.array([0.0, 2.0])

    def args(self):
        return None

    def expected_result(self):
        # The solution is at x1 = 1, x2 = 1
        return jnp.array([1.0, 1.0])

    def expected_objective_value(self):
        # Evaluating at (1, 1): 1³ - 3*1 + 1² - 2*1 + 2 = 1 - 3 + 1 - 2 + 2 = -1
        return jnp.array(-1.0)



import jax
import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class ECKERLE4LS(AbstractUnconstrainedMinimisation):
    """ECKERLE4LS - Nonlinear Least-Squares problem (NIST dataset).

    This problem involves a nonlinear least squares fit to a Gaussian peak function,
    arising from a circular interference transmittance study.

    Source: Problem 7 from
    NIST nonlinear least squares test set
    http://www.itl.nist.gov/div898/strd/nls/nls_main.shtml

    SIF input: Ph. Toint, April 1997.

    Classification: SUR2-MN-3-0
    """

    # Allow selecting which starting point to use (0-based indexing)
    start_point: int = 0  # 0 or 1
    _allowed_start_points = frozenset({0, 1})

    def __check_init__(self):
        if self.start_point not in self._allowed_start_points:
            allowed = self._allowed_start_points
            msg = f"start_point must be in {allowed}, got {self.start_point}"
            raise ValueError(msg)

    def objective(self, y, args):
        del args

        # Extract parameters
        b1, b2, b3 = y

        # NIST dataset: x and y values
        x_data = jnp.array(
            [
                400.0,
                405.0,
                410.0,
                415.0,
                420.0,
                425.0,
                430.0,
                435.0,
                436.5,
                438.0,
                439.5,
                441.0,
                442.5,
                444.0,
                445.5,
                447.0,
                448.5,
                450.0,
                451.5,
                453.0,
                454.5,
                456.0,
                457.5,
                459.0,
                460.5,
                462.0,
                463.5,
                465.0,
                470.0,
                475.0,
                480.0,
                485.0,
                490.0,
                495.0,
                500.0,
            ]
        )

        y_data = jnp.array(
            [
                1.575e-1,
                1.699e-1,
                2.350e-1,
                3.102e-1,
                4.917e-1,
                8.710e-1,
                1.718e0,
                3.682e0,
                4.944e0,
                6.637e0,
                8.796e0,
                1.168e1,
                1.484e1,
                1.739e1,
                1.710e1,
                1.342e1,
                7.600e0,
                2.245e0,
                1.120e0,
                8.178e-1,
                6.615e-1,
                5.880e-1,
                4.401e-1,
                3.431e-1,
                2.400e-1,
                1.909e-1,
                1.750e-1,
                1.381e-1,
                9.576e-2,
                7.274e-2,
                6.907e-2,
                6.012e-2,
                5.629e-2,
                4.992e-2,
                4.670e-2,
            ]
        )

        # Model function: b1/b2 * exp[-0.5*((x-b3)/b2)^2]
        def model_func(x):
            exponent = -0.5 * ((x - b3) / b2) ** 2
            return (b1 / b2) * jnp.exp(exponent)

        # Calculate model predictions for all x values at once
        predictions = jax.vmap(model_func)(x_data)

        # Compute residuals
        residuals = predictions - y_data

        # Return the sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Two starting points from the SIF file
        if self.start_point == 0:
            # Starting point 1: b1 = 1.0, b2 = 10.0, b3 = 500.0
            return jnp.array([1.0, 10.0, 500.0])
        else:  # self.start_point == 1
            # Starting point 2: b1 = 1.5, b2 = 5.0, b3 = 450.0
            return jnp.array([1.5, 5.0, 450.0])

    def args(self):
        return None

    def expected_result(self):
        # Should use certified values from NIST
        return None

    def expected_objective_value(self):
        # Should use certified minimum value from NIST
        return None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HS9(AbstractConstrainedMinimisation):
    """Problem 9 from the Hock-Schittkowski test collection.

    A 2-variable trigonometric function with one equality constraint.

    f(x) = sin(πx₁/12) cos(πx₂/16)

    Subject to: 4x₁ - 3x₂ = 0

    Source: problem 9 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    Also cited: Miele e.al. [44]

    Classification: GLR-T1-1
    """

    def objective(self, y, args):
        x1, x2 = y
        return jnp.sin(jnp.pi * x1 / 12) * jnp.cos(jnp.pi * x2 / 16)

    def y0(self):
        return jnp.array([0.0, 0.0])

    def args(self):
        return None

    def expected_result(self):
        # From PDF: x* = (12k - 3, 16k - 4), k=0, ±1, ±2, ...
        # Using k=-1 to get the solution that gives f(x*) = -0.5
        return jnp.array([-15.0, -20.0])

    def expected_objective_value(self):
        return jnp.array(-0.5)

    def bounds(self):
        return None

    def constraint(self, y):
        x1, x2 = y
        # Equality constraint: 4x₁ - 3x₂ = 0
        equality_constraint = 4 * x1 - 3 * x2
        return equality_constraint, None



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: Human review needed to verify the implementation matches the problem definition
class JUDGE(AbstractUnconstrainedMinimisation):
    """SCIPY global optimization benchmark example Judge.

    Fit: y = x_1 + a_i * x_2 + b_i * x_2^2 + e

    Source: Problem from the SCIPY benchmark set
    https://github.com/scipy/scipy/tree/master/benchmarks/benchmarks/go_benchmark_functions

    SIF input: Nick Gould, Jan 2020
    Classification: SUR2-MN-2-0
    """

    def objective(self, y, args):
        """Compute the objective function value.

        For each data point i, the residual is:
        x_1 + a_i * x_2 + b_i * x_2^2 - y_i

        The objective is the sum of squares of these residuals.
        """
        x1, x2 = y

        # Calculate the predicted values using the model
        # y_pred = x_1 + a_i * x_2 + (b_i^2) * x_2
        a_values = jnp.array(
            [
                0.286,
                0.973,
                0.384,
                0.276,
                0.973,
                0.543,
                0.957,
                0.948,
                0.543,
                0.797,
                0.936,
                0.889,
                0.006,
                0.828,
                0.399,
                0.617,
                0.939,
                0.784,
                0.072,
                0.889,
            ]
        )
        b_values = jnp.array(
            [
                0.645,
                0.585,
                0.310,
                0.058,
                0.455,
                0.779,
                0.259,
                0.202,
                0.028,
                0.099,
                0.142,
                0.296,
                0.175,
                0.180,
                0.842,
                0.039,
                0.103,
                0.620,
                0.158,
                0.704,
            ]
        )
        y_values = jnp.array(
            [
                4.284,
                4.149,
                3.877,
                0.533,
                2.211,
                2.389,
                2.145,
                3.231,
                1.998,
                1.379,
                2.106,
                1.428,
                1.011,
                2.179,
                2.858,
                1.388,
                1.651,
                1.593,
                1.046,
                2.152,
            ]
        )

        # Calculate predicted values: x1 + a_i * x2 + b_i * x2^2
        y_pred = x1 + a_values * x2 + b_values * (x2**2)

        # Calculate the residuals
        residuals = y_pred - y_values

        # Return the sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        """Initial point (1.0, 5.0)."""
        return jnp.array([1.0, 5.0])

    def args(self):
        """No additional arguments needed."""
        return None

    def expected_result(self):
        """The solution is not specified in the SIF file."""
        return None

    def expected_objective_value(self):
        """The solution value seems to be 0.0 according to the SIF file comment."""
        return jnp.array(0.0)



import jax
import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: This implementation requires human review and verification against
# another CUTEst interface
class BIGGS6(AbstractUnconstrainedMinimisation):
    """Biggs EXP problem in 6 variables.

    Source: Problem 21 in
    A.R. Buckley,
    "Test functions for unconstrained minimization",
    TR 1989CS-3, Mathematics, statistics and computing centre,
    Dalhousie University, Halifax (CDN), 1989.

    SIF input: Ph. Toint, Dec 1989.

    Classification: SUR2-AN-6-0
    """

    n: int = 6  # Problem has 6 variables
    m: int = 13  # Number of data points

    def objective(self, y, args):
        del args
        x1, x2, x3, x4, x5, x6 = y

        # Define indices from 1 to m
        indices = jnp.arange(1, self.m + 1)

        # Define inner function to compute residual for a single index i
        def compute_residual(i):
            t = -0.1 * i

            # Target value calculation from the SIF file
            y_val = jnp.exp(t) - 5.0 * jnp.exp(-1.0 * i) + 3.0 * jnp.exp(4.0 * t)

            # Model calculation
            y_pred = x3 * jnp.exp(t * x1) - x4 * jnp.exp(t * x2) + x6 * jnp.exp(t * x5)

            # Return squared residual
            return (y_pred - y_val) ** 2

        # Vectorize the function over indices and sum the results
        residuals = jax.vmap(compute_residual)(indices)
        return jnp.sum(residuals)

    def y0(self):
        # Initial values from SIF file
        return jnp.array([1.0, 2.0, 1.0, 1.0, 1.0, 1.0])

    def args(self):
        return None

    def expected_result(self):
        # The optimal solution is known to be (1, 10, 1, 5, 4, 3)
        # per other references, including Buckley's original paper
        return jnp.array([1.0, 10.0, 1.0, 5.0, 4.0, 3.0])

    def expected_objective_value(self):
        # According to the SIF file comment (line 135),
        # the optimal objective value is 0.0
        return jnp.array(0.0)



from .bt1 import BT1 as BT1
from .bt2 import BT2 as BT2
from .bt3 import BT3 as BT3
from .bt4 import BT4 as BT4
from .bt5 import BT5 as BT5
from .bt6 import BT6 as BT6
from .bt7 import BT7 as BT7
from .bt8 import BT8 as BT8
from .bt9 import BT9 as BT9
from .bt10 import BT10 as BT10
from .bt11 import BT11 as BT11
from .bt12 import BT12 as BT12
from .bt13 import BT13 as BT13
from .hs6 import HS6 as HS6
from .hs7 import HS7 as HS7
from .hs8 import HS8 as HS8
from .hs9 import HS9 as HS9
from .hs10 import HS10 as HS10
from .hs11 import HS11 as HS11
from .hs12 import HS12 as HS12
from .hs13 import HS13 as HS13
from .hs14 import HS14 as HS14
from .hs15 import HS15 as HS15
from .hs16 import HS16 as HS16
from .hs17 import HS17 as HS17
from .hs18 import HS18 as HS18
from .hs19 import HS19 as HS19
from .hs20 import HS20 as HS20
from .hs21 import HS21 as HS21
from .hs22 import HS22 as HS22
from .hs23 import HS23 as HS23
from .hs24 import HS24 as HS24
from .hs26 import HS26 as HS26
from .hs27 import HS27 as HS27
from .hs28 import HS28 as HS28
from .hs29 import HS29 as HS29
from .hs30 import HS30 as HS30
from .hs31 import HS31 as HS31
from .hs32 import HS32 as HS32
from .hs33 import HS33 as HS33
from .hs34 import HS34 as HS34
from .hs35 import HS35 as HS35
from .hs36 import HS36 as HS36
from .hs37 import HS37 as HS37
from .hs39 import HS39 as HS39
from .hs40 import HS40 as HS40
from .hs41 import HS41 as HS41
from .hs42 import HS42 as HS42
from .hs43 import HS43 as HS43
from .hs44 import HS44 as HS44
from .hs46 import HS46 as HS46
from .hs47 import HS47 as HS47
from .hs48 import HS48 as HS48
from .hs49 import HS49 as HS49
from .hs50 import HS50 as HS50
from .hs51 import HS51 as HS51
from .hs52 import HS52 as HS52
from .hs53 import HS53 as HS53
from .hs54 import HS54 as HS54
from .hs55 import HS55 as HS55
from .hs56 import HS56 as HS56
from .hs57 import HS57 as HS57

# from .hs59 import HS59 as HS59  # TODO: Human review - objective function discrepancy
from .hs60 import HS60 as HS60
from .hs61 import HS61 as HS61
from .hs62 import HS62 as HS62
from .hs63 import HS63 as HS63
from .hs64 import HS64 as HS64
from .hs65 import HS65 as HS65
from .hs66 import HS66 as HS66

# from .hs67 import HS67 as HS67  # TODO: Human review - several discrepancies
from .hs68 import HS68 as HS68
from .hs69 import HS69 as HS69
from .hs71 import HS71 as HS71
from .hs72 import HS72 as HS72
from .hs73 import HS73 as HS73
from .hs74 import HS74 as HS74
from .hs75 import HS75 as HS75
from .hs76 import HS76 as HS76
from .hs77 import HS77 as HS77
from .hs78 import HS78 as HS78
from .hs79 import HS79 as HS79
from .hs80 import HS80 as HS80
from .hs81 import HS81 as HS81
from .hs83 import HS83 as HS83
from .hs93 import HS93 as HS93

# from .hs99 import HS99 as HS99  # TODO: Needs human review - complex recursion
from .hs100 import HS100 as HS100
from .hs101 import HS101 as HS101
from .hs102 import HS102 as HS102
from .hs103 import HS103 as HS103
from .hs104 import HS104 as HS104
from .hs105 import HS105 as HS105
from .hs106 import HS106 as HS106
from .hs107 import HS107 as HS107
from .hs108 import HS108 as HS108
from .hs109 import HS109 as HS109
from .hs111 import HS111 as HS111
from .hs112 import HS112 as HS112
from .hs113 import HS113 as HS113
from .hs114 import HS114 as HS114
from .hs116 import HS116 as HS116
from .hs117 import HS117 as HS117
from .hs118 import HS118 as HS118
from .hs119 import HS119 as HS119
from .lukvle1 import LUKVLE1 as LUKVLE1

# from .lukvle2 import LUKVLE2 as LUKVLE2
from .lukvle3 import LUKVLE3 as LUKVLE3

# from .lukvle4 import LUKVLE4 as LUKVLE4  # Use LUKVLE4C instead
# from .lukvle4c import LUKVLE4C as LUKVLE4C
from .lukvle5 import LUKVLE5 as LUKVLE5
from .lukvle6 import LUKVLE6 as LUKVLE6
from .lukvle7 import LUKVLE7 as LUKVLE7
from .lukvle8 import LUKVLE8 as LUKVLE8
from .lukvle9 import LUKVLE9 as LUKVLE9
from .lukvle10 import LUKVLE10 as LUKVLE10
from .lukvle11 import LUKVLE11 as LUKVLE11

# from .lukvle12 import LUKVLE12 as LUKVLE12  # Has constraint function inconsistencies
from .lukvle13 import LUKVLE13 as LUKVLE13
from .lukvle14 import LUKVLE14 as LUKVLE14
from .lukvle15 import LUKVLE15 as LUKVLE15
from .lukvle16 import LUKVLE16 as LUKVLE16
from .lukvle17 import LUKVLE17 as LUKVLE17
from .lukvle18 import LUKVLE18 as LUKVLE18
from .lukvli1 import LUKVLI1 as LUKVLI1

# from .lukvli2 import LUKVLI2 as LUKVLI2
from .lukvli3 import LUKVLI3 as LUKVLI3

# from .lukvli4 import LUKVLI4 as LUKVLI4  # Use LUKVLI4C instead
# from .lukvli4c import LUKVLI4C as LUKVLI4C
from .lukvli5 import LUKVLI5 as LUKVLI5
from .lukvli6 import LUKVLI6 as LUKVLI6
from .lukvli7 import LUKVLI7 as LUKVLI7
from .lukvli8 import LUKVLI8 as LUKVLI8
from .lukvli9 import LUKVLI9 as LUKVLI9
from .lukvli10 import LUKVLI10 as LUKVLI10
from .lukvli11 import LUKVLI11 as LUKVLI11

# from .lukvli12 import LUKVLI12 as LUKVLI12  # Has constraint function inconsistencies
from .lukvli13 import LUKVLI13 as LUKVLI13
from .lukvli14 import LUKVLI14 as LUKVLI14
from .lukvli15 import LUKVLI15 as LUKVLI15
from .lukvli16 import LUKVLI16 as LUKVLI16
from .lukvli17 import LUKVLI17 as LUKVLI17
from .lukvli18 import LUKVLI18 as LUKVLI18


constrained_minimisation_problems = (
    HS6(),
    HS7(),
    HS8(),
    HS9(),
    HS10(),
    HS11(),
    HS12(),
    HS13(),
    HS14(),
    HS15(),
    HS16(),
    HS17(),
    HS18(),
    HS19(),
    HS20(),
    HS21(),
    HS22(),
    HS23(),
    HS24(),
    HS26(),
    HS27(),
    HS28(),
    HS29(),
    HS30(),
    HS31(),
    HS32(),
    HS33(),
    HS34(),
    HS35(),
    HS36(),
    HS37(),
    HS39(),
    HS40(),
    HS41(),
    HS42(),
    HS43(),
    HS44(),
    HS46(),
    HS47(),
    HS48(),
    HS49(),
    HS50(),
    HS51(),
    HS52(),
    HS53(),
    HS54(),
    HS55(),
    HS56(),
    HS57(),
    # HS59(),  # TODO: Human review - objective function discrepancy
    HS60(),
    HS61(),
    HS62(),
    HS63(),
    HS64(),
    HS65(),
    HS66(),
    # HS67(),  # TODO: Human review - several discrepancies
    HS68(),
    HS69(),
    HS71(),
    HS72(),
    HS73(),
    HS74(),
    HS75(),
    HS76(),
    HS77(),
    HS78(),
    HS79(),
    HS80(),
    HS81(),
    HS83(),
    HS93(),
    # HS99(),  # TODO: Needs human review - complex recursive formulation
    HS100(),
    HS101(),
    HS102(),
    HS103(),
    HS104(),
    HS105(),
    HS106(),
    HS107(),
    HS108(),
    HS109(),
    HS111(),
    HS112(),
    HS113(),
    HS114(),
    HS116(),
    HS117(),
    HS118(),
    HS119(),
    BT1(),
    BT2(),
    BT3(),
    BT4(),
    BT5(),
    BT6(),
    BT7(),
    BT8(),
    BT9(),
    BT10(),
    BT11(),
    BT12(),
    BT13(),
    LUKVLE1(),
    # LUKVLE2(),
    LUKVLE3(),
    # LUKVLE4(),  # Has factor ~2380 error due to SIF bug, use LUKVLE4C instead
    # LUKVLE4C(),  # Has 3% discrepancy with pycutest
    LUKVLE5(),
    LUKVLE6(),
    LUKVLE7(),
    LUKVLE8(),
    LUKVLE9(),
    LUKVLE10(),
    LUKVLE11(),
    # LUKVLE12(),  # Has constraint function inconsistencies
    LUKVLE13(),
    LUKVLE14(),
    LUKVLE15(),
    LUKVLE16(),
    LUKVLE17(),
    LUKVLE18(),
    LUKVLI1(),
    # LUKVLI2(),
    LUKVLI3(),
    # LUKVLI4(),  # Has factor ~2380 error due to SIF bug, use LUKVLI4C instead
    # LUKVLI4C(),  # Has 3% discrepancy with pycutest
    LUKVLI5(),
    LUKVLI6(),
    LUKVLI7(),
    LUKVLI8(),
    LUKVLI9(),
    LUKVLI10(),
    LUKVLI11(),
    # LUKVLI12(),  # Has constraint function inconsistencies
    LUKVLI13(),
    LUKVLI14(),
    LUKVLI15(),
    LUKVLI16(),
    LUKVLI17(),
    LUKVLI18(),
)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class BT4(AbstractConstrainedMinimisation):
    """BT4 - Boggs-Tolle test problem 4.

    n = 3, m = 2.
    f(x) = x₁ - x₂ + x₂³.
    g₁(x) = x₁² + x₂² + x₃² - 25.
    g₂(x) = x₁ + x₂ + x₃ - 1.

    Start 1: x₁ = 3.1494, x₂ = 1.4523, x₃ = -3.6017.
    Start 2: x₁ = 3.122, x₂ = 1.489, x₃ = -3.611.
    Start 3: x₁ = -0.94562, x₂ = -2.35984, x₃ = 4.30546.
    Solution: x* = (4.0382, -2.9470, -0.09115).

    Source: Boggs, P.T. and Tolle, J.W.,
    "A strategy for global convergence in a sequential
    quadratic programming algorithm",
    SIAM J. Numer. Anal. 26(3), pp. 600-623, 1989.

    Classification: QQR2-AN-3-2
    """

    y0_id: int = 3  # Default to solution from paper as per SIF
    provided_y0s: frozenset = frozenset({0, 1, 2, 3})

    def objective(self, y, args):
        del args
        x1, x2, x3 = y
        return x1 - x2 + x2**3

    def y0(self):
        if self.y0_id == 0:
            return jnp.array([3.1494, 1.4523, -3.6017])
        elif self.y0_id == 1:
            return jnp.array([3.122, 1.489, -3.611])
        elif self.y0_id == 2:
            return jnp.array([-0.94562, -2.35984, 4.30546])
        elif self.y0_id == 3:
            return jnp.array([4.0382, -2.9470, -0.09115])

    def args(self):
        return None

    def expected_result(self):
        return jnp.array([4.0382, -2.9470, -0.09115])

    def expected_objective_value(self):
        return None  # Not explicitly given

    def bounds(self):
        return None

    def constraint(self, y):
        x1, x2, x3 = y
        # Equality constraints
        g1 = x1**2 + x2**2 + x3**2 - 25
        g2 = x1 + x2 + x3 - 1
        equality_constraints = jnp.array([g1, g2])
        return equality_constraints, None



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class BROYDN3DLS(AbstractUnconstrainedMinimisation):
    """Broyden tridiagonal system of nonlinear equations in the least square sense.

    Source: problem 30 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Toint#17 and Buckley#78.
    SIF input: Ph. Toint, Dec 1989.
    Least-squares version: Nick Gould, Oct 2015.

    Classification: SUR2-AN-V-0
    """

    n: int = 5000  # Dimension of the problem
    kappa1: float = 2.0  # Parameter
    kappa2: float = 1.0  # Parameter

    def objective(self, y, args):
        del args
        n = self.n
        k1 = self.kappa1
        k2 = self.kappa2

        # Compute residuals without conditional logic
        residuals = jnp.zeros(n)

        # First residual: (3-2*x1)*x1 - 2*x2 + k2
        residuals = residuals.at[0].set((3.0 - k1 * y[0]) * y[0] - 2.0 * y[1] + k2)

        # Last residual: (3-2*xn)*xn - xn-1 + k2
        residuals = residuals.at[n - 1].set(
            (3.0 - k1 * y[n - 1]) * y[n - 1] - y[n - 2] + k2
        )

        # Middle residuals: (3-2*xi)*xi - xi-1 - 2*xi+1 + k2 for i=1 to n-2
        if n > 2:
            middle_indices = jnp.arange(1, n - 1)
            middle_residuals = (
                (3.0 - k1 * y[middle_indices]) * y[middle_indices]
                - y[middle_indices - 1]
                - 2.0 * y[middle_indices + 1]
                + k2
            )
            residuals = residuals.at[middle_indices].set(middle_residuals)

        # Return the sum of squared residuals
        return jnp.sum(residuals**2)

    def y0(self):
        # Initial values from SIF file (all -1.0)
        return jnp.full(self.n, -1.0)

    def args(self):
        return None

    def expected_result(self):
        # Set values of all components to the same value r
        # where r is approximately -k2/(n*k1)
        k2 = self.kappa2
        n = self.n
        k1 = self.kappa1
        r = -k2 / (n * k1)
        return jnp.full(self.n, r)

    def expected_objective_value(self):
        # According to the SIF file comment (line 110),
        # the optimal objective value is 0.0
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class LUKVLI2(AbstractConstrainedMinimisation):
    """LUKVLI2 - Chained Wood function with Broyden banded inequality constraints.

    Problem 5.2 from Luksan and Vlcek test problems with inequality constraints.

    The objective is a chained Wood function:
    f(x) = Σ[i=1 to n/2-1] [100(x_{2i-1}^2 - x_{2i})^2 + (x_{2i-1} - 1)^2 +
                             90(x_{2i+1}^2 - x_{2i+2})^2 + (x_{2i+1} - 1)^2 +
                             10(x_{2i} + x_{2i+2} - 2)^2 + (x_{2i} - x_{2i+2})^2/10]

    Subject to inequality constraints:
    c_k(x) = (2 + 5x_{k+5}^2)x_{k+5} + 1 + Σ[i=k-5 to k+1] x_i(1 + x_i) ≤ 0,
    for k = 6, ..., n-2 (with appropriate adjustments for boundary cases)

    Starting point: x_i = -2 for i odd, x_i = 1 for i even

    Source: L. Luksan and J. Vlcek,
    "Sparse and partially separable test problems for
    unconstrained and equality constrained optimization",
    Technical Report 767, Inst. Computer Science, Academy of Sciences
    of the Czech Republic, 182 07 Prague, Czech Republic, 1999

    Equality constraints changed to inequalities

    SIF input: Nick Gould, April 2001

    Classification: OOR2-AY-V-V
    """

    n: int = 10000  # Default dimension, can be overridden

    def objective(self, y, args):
        # TODO: This implementation matches the structure in the SIF file but
        # differs from pycutest by a factor of ~34.5. The SIF file uses
        # group-separable structure where scale factors are applied in the
        # denominator, but even accounting for this there is a discrepancy.
        # This needs further investigation.
        del args
        n = len(y)
        # Chained Wood function - vectorized
        num_terms = n // 2 - 1
        if num_terms <= 0:
            return jnp.array(0.0)

        # Extract all required elements at once
        i = jnp.arange(num_terms)
        x_2i_minus_1 = y[2 * i]  # x_{2i-1} in 1-based
        x_2i = y[2 * i + 1]  # x_{2i} in 1-based
        x_2i_plus_1 = y[2 * i + 2]  # x_{2i+1} in 1-based
        x_2i_plus_2 = y[2 * i + 3]  # x_{2i+2} in 1-based

        # Compute all terms vectorized
        terms = (
            100 * (x_2i_minus_1**2 - x_2i) ** 2
            + (x_2i_minus_1 - 1) ** 2
            + 90 * (x_2i_plus_1**2 - x_2i_plus_2) ** 2
            + (x_2i_plus_1 - 1) ** 2
            + 10 * (x_2i + x_2i_plus_2 - 2) ** 2
            + (x_2i - x_2i_plus_2) ** 2 / 10
        )
        return jnp.sum(terms)

    def y0(self):
        # Starting point: x_i = -2 for i odd, x_i = 1 for i even
        y = jnp.zeros(self.n)
        # JAX uses 0-based indexing, so odd indices in the problem are even in JAX
        y = y.at[::2].set(-2.0)  # i = 1, 3, 5, ... (1-based) -> 0, 2, 4, ... (0-based)
        y = y.at[1::2].set(1.0)  # i = 2, 4, 6, ... (1-based) -> 1, 3, 5, ... (0-based)
        return y

    def args(self):
        return None

    def expected_result(self):
        # Solution is all ones
        return jnp.ones(self.n)

    def expected_objective_value(self):
        return jnp.array(0.0)

    def bounds(self):
        return None

    def constraint(self, y):
        n = len(y)
        if n < 10:  # Need at least 10 elements for meaningful constraints
            return None, jnp.array([])

        # Inequality constraints - vectorized
        # Constraints from k=6 to n-2 (1-based), which is k=5 to n-3 (0-based)
        k_values = jnp.arange(5, n - 2)
        num_constraints = len(k_values)

        if num_constraints == 0:
            return None, jnp.array([])

        # k+5 in 1-based is k+4 in 0-based
        k_plus_5_idx = k_values + 4

        # (2 + 5x_{k+5}^2)x_{k+5} + 1 for all constraints
        main_terms = (2 + 5 * y[k_plus_5_idx] ** 2) * y[k_plus_5_idx] + 1

        # For the sum, we need x_i(1 + x_i) for i in [max(0, k-5), min(n-1, k+1)]
        # This is similar to the window sum in LUKVLI6
        # Pre-compute x_i(1 + x_i) for all elements
        x_terms = y * (1 + y)

        # Use cumulative sum approach for efficient window computation
        # Pad x_terms for easier indexing
        padded_x_terms = jnp.pad(x_terms, (6, 1), mode="constant")
        cumsum = jnp.cumsum(padded_x_terms)

        # For each k, we want sum from max(0, k-5) to min(n-1, k+1)
        # In padded array, position i corresponds to i+6
        # For k in original array, we want sum from (k-5) to (k+1) inclusive
        # In padded: from (k-5+6) to (k+1+6) = from (k+1) to (k+7)
        # This is cumsum[k+7] - cumsum[k]
        # But we need to adjust for the actual k_values which start at 5
        start_positions = k_values + 1
        end_positions = k_values + 7

        window_sums = cumsum[end_positions] - cumsum[start_positions - 1]

        inequality_constraints = main_terms + window_sums
        return None, inequality_constraints



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: This implementation requires human review and verification against
# another CUTEst interface
class BROWNAL(AbstractUnconstrainedMinimisation):
    """Brown almost linear least squares problem.

    This problem is a sum of n least-squares groups, the last one of
    which has a nonlinear element. Its Hessian matrix is dense.

    Source: Problem 27 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley#79
    SIF input: Ph. Toint, Dec 1989.

    Classification: SUR2-AN-V-0
    """

    n: int = 200  # Other suggested dimensions are 10, 100, and 1000

    def objective(self, y, args):
        del args
        n = self.n

        # Sum of squared residuals
        residuals = []

        # First n-1 groups: each group G(i) has 2*x[i] + sum_{j≠i}(x[j]) - (n+1)
        # This is equivalent to x[i] + sum_all(x[j]) - (n+1)
        sum_all = jnp.sum(y)
        for i in range(n - 1):
            residual = y[i] + sum_all - (n + 1)
            residuals.append(residual)

        # Last group: product of components - 1
        product = jnp.prod(y)
        residuals.append(product - 1.0)

        return jnp.sum(jnp.array(residuals) ** 2)

    def y0(self):
        # Initial values from SIF file (all 0.5)
        return jnp.full(self.n, 0.5)

    def args(self):
        return None

    def expected_result(self):
        # The optimal solution for Brown almost linear problem is all variables = 1
        # This satisfies both constraints:
        # - Linear constraints: x[i] + sum(x) - (n+1) = 1 + n - (n+1) = 0
        # - Product constraint: prod(x) - 1 = 1 - 1 = 0
        # giving an optimal objective value of 0.0
        return jnp.ones(self.n)

    def expected_objective_value(self):
        # The optimal objective value for Brown almost linear problem is 0.0
        # when all variables equal 1. This satisfies both:
        # - Linear constraints: x[i] + sum(x) - (n+1) = 1 + n - (n+1) = 0
        # - Product constraint: prod(x) - 1 = 1 - 1 = 0
        return jnp.array(0.0)
